#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests the edit views."""

from wikkid.tests.factory import ViewTestCase
from wikkid.tests.fakes import TestUser


class TestEdit(ViewTestCase):
    """Test the edit view."""

    def setUp(self):
        super(TestEdit, self).setUp()
        self.user = TestUser('test@example.com', 'Test User')

    def test_title_nested(self):
        """Test that a nested page returns the expected title"""
        factory = self.make_factory([
                ('SomePage/SubPage/Nested.txt', 'some text')])
        view = self.get_view(factory, '/SomePage/SubPage', 'edit')
        self.assertEqual('Editing "Sub Page"', view.title)

    def test_urls(self):
        """Check the urls for saving and cancel."""
        factory = self.make_factory()
        view = self.get_view(factory, '/NewPage', 'edit')
        self.assertEqual('/NewPage', view.cancel_url)
        self.assertEqual('/NewPage/+save', view.save_url)



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The base user implementation.

Provides the gravatar support.
"""

import hashlib


class BaseUser(object):

    @property
    def gravatar(self):
        url = "http://www.gravatar.com/avatar/"
        data = self.email.lower().encode('utf-8', 'replace')
        url += hashlib.md5(data).hexdigest()
        url += "?s=50&d=identicon"
        return url



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Text to HTML formatters.

The initial implemention is ReST, but I'm hopeful that there will soon be
pygments, and other wiki text formatters.
"""



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A volatile filestore.

Used primarily for test purposes, this class should be a fully functional
filestore, albiet one that doesn't remember anything persistently.
"""

from datetime import datetime
from itertools import count

from breezy.urlutils import dirname
from zope.interface import implementer

from wikkid.filestore import FileExists
from wikkid.filestore.basefile import BaseFile
from wikkid.interface.filestore import FileType, IFile, IFileStore


@implementer(IFileStore)
class FileStore(object):
    """A filestore that just uses an internal map to store data."""

    def __init__(self, files=None):
        """Files is a list of tuples.

        If the content is None, the path is assumed to be a directory.  If the
        content contains a null character, the file is considered binary.
        """
        self._integer = count(1)
        self.file_id_map = {}
        self.path_map = {}
        if files is None:
            files = []
        user = 'First User <first@example.com'
        for path, content in files:
            if path.endswith('/'):
                self._ensure_dir(path[:-1], user)
            else:
                self._add_file(path, content, user)

    def _ensure_dir(self, path, user):
        # If we are at the start, we are done.
        if not path:
            return
        # If the directory exists, we are done.
        if path in self.path_map:
            # Check to make sure that it is a directory.
            if self.path_map[path].file_type != FileType.DIRECTORY:
                raise FileExists(
                    "Found a file at '%s' where a directory is needed"
                    % path)
            return
        # Check to make sure the parent is in there too.
        self._ensure_dir(dirname(path), user)
        new_dir = File(path, None, next(self._integer), user)
        self.file_id_map[new_dir.file_id] = new_dir
        self.path_map[new_dir.path] = new_dir

    def _add_file(self, path, content, user):
        self._ensure_dir(dirname(path), user)
        new_file = File(path, content, next(self._integer), user)
        self.file_id_map[new_file.file_id] = new_file
        self.path_map[new_file.path] = new_file

    def get_file(self, path):
        """Return an object representing the file."""
        if path in self.path_map:
            return self.path_map[path]
        else:
            return None

    def update_file(self, path, content, user, parent_revision,
                    commit_message=None):
        """The `user` is updating the file at `path` with `content`."""
        existing_file = self.get_file(path)
        if existing_file is None:
            self._add_file(path, content, user)
        else:
            existing_file.content = content
            existing_file.last_modified_by = user
            existing_file.last_modified_date = datetime.utcnow()

    def list_directory(self, directory_path):
        """Return a list of File objects for in the directory path.

        If the path doesn't exist, returns None.  If the path exists but is
        empty, an empty list is returned.  Otherwise a list of File objects in
        that directory.
        """
        if directory_path is None:
            directory_path = ''
        else:
            directory = self.get_file(directory_path)
            if directory is None or not directory._is_directory:
                return None
        listing = []
        for path, value in self.path_map.items():
            path_dir = dirname(path)
            if path_dir == directory_path:
                listing.append(value)
        return listing


@implementer(IFile)
class File(BaseFile):
    """A volatile file object."""

    def __init__(self, path, content, file_id, user):
        BaseFile.__init__(self, path)
        self.file_id = file_id
        self.content = content
        self.last_modified_in_revision = None
        self.last_modified_by = user
        self.last_modified_date = datetime.utcnow()
        self.file_type = self._get_filetype()

    def _get_filetype(self):
        """Work out the filetype based on the mimetype if possible."""
        if self._is_directory:
            return FileType.DIRECTORY
        else:
            if self._mimetype is None:
                binary = self._is_binary
            else:
                binary = not self._mimetype.startswith('text/')
            if binary:
                return FileType.BINARY_FILE
            else:
                return FileType.TEXT_FILE

    def get_content(self):
        return self.content

    @property
    def _is_directory(self):
        return self.content is None

    @property
    def _is_binary(self):
        return b'\0' in self.content



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The base test class for filestores."""

from datetime import datetime, timedelta

from wikkid.filestore import FileExists
from wikkid.interface.filestore import FileType, IFile, IFileStore


class TestFileStore(object):
    """Tests for the filestore and files."""

    def test_filestore_provides_IFileStore(self):
        filestore = self.make_filestore()
        self.assertProvides(filestore, IFileStore)

    def test_file_provides_IFile(self):
        filestore = self.make_filestore([('README', b'not much')])
        readme = filestore.get_file('README')
        self.assertProvides(readme, IFile)

    def test_file_gives_content(self):
        filestore = self.make_filestore([('README', b'Content')])
        readme = filestore.get_file('README')
        self.assertEqual(b'Content', readme.get_content())

    def assertDirectoryFileType(self, f):
        self.assertEqual(FileType.DIRECTORY, f.file_type)

    def assertTextFileType(self, f):
        self.assertEqual(FileType.TEXT_FILE, f.file_type)

    def assertBinaryFileType(self, f):
        self.assertEqual(FileType.BINARY_FILE, f.file_type)

    def test_file_type(self):
        filestore = self.make_filestore(
            [('README', b'Content'),
             ('lib/', None),
             ('lib/foo', b'dummy data'),
             ('image.jpg', b'pretend image'),
             ('binary-file', b'a\0binary\0file'),
             ('simple.txt', b'A text file'),
             ('source.cpp', b'A cpp file')])
        self.assertDirectoryFileType(filestore.get_file('lib'))
        self.assertTextFileType(filestore.get_file('README'))
        self.assertTextFileType(filestore.get_file('simple.txt'))
        self.assertTextFileType(filestore.get_file('source.cpp'))
        self.assertBinaryFileType(filestore.get_file('image.jpg'))
        self.assertBinaryFileType(filestore.get_file('binary-file'))

    def test_mimetype(self):
        filestore = self.make_filestore(
            [('README', b'Content'),
             ('lib/', None),
             ('lib/data', b'dummy data'),
             ('image.jpg', b'pretend image'),
             ('binary-file', b'a\0binary\0file'),
             ('simple.txt', b'A text file'),
             ('source.cpp', b'A cpp file')])
        self.assertIs(None, filestore.get_file('lib').mimetype)
        self.assertIs(None, filestore.get_file('README').mimetype)
        self.assertEqual(
            'text/plain', filestore.get_file('simple.txt').mimetype)
        self.assertEqual(
            'text/x-c++src', filestore.get_file('source.cpp').mimetype)
        self.assertEqual(
            'image/jpeg', filestore.get_file('image.jpg').mimetype)
        self.assertIs(None, filestore.get_file('binary-file').mimetype)

    def test_nonexistant_file(self):
        filestore = self.make_filestore()
        readme = filestore.get_file('README')
        self.assertIs(None, readme)

    def assertDirectory(self, filestore, path):
        """The filestore should have a directory at path."""
        location = filestore.get_file(path)
        self.assertDirectoryFileType(location)

    def test_updating_file_adds_directories(self):
        filestore = self.make_filestore()
        user = 'Eric the viking <eric@example.com>'
        filestore.update_file('first/second/third', b'content\n', user,
                              None)
        self.assertDirectory(filestore, 'first')
        self.assertDirectory(filestore, 'first/second')
        third = filestore.get_file('first/second/third')
        self.assertEqual(b'content\n', third.get_content())

    def test_updating_file_with_directory_clash(self):
        filestore = self.make_filestore(
            [('first', b'content')])
        user = None
        self.assertRaises(
            FileExists, filestore.update_file,
            'first/second', b'content', user, None)

    def test_updating_existing_file(self):
        filestore = self.make_filestore(
            [('README', b'Content'),
             ])
        user = 'Eric the viking <eric@example.com>'
        parent_rev = filestore.get_file('README').last_modified_in_revision
        filestore.update_file('README', b'new content\n', user, parent_rev)
        readme = filestore.get_file('README')
        self.assertEqual(b'new content\n', readme.get_content())

    def test_list_directory_non_existant(self):
        filestore = self.make_filestore()
        listing = filestore.list_directory('missing')
        self.assertIs(None, listing)

    def test_listing_directory_empty(self):
        filestore = self.make_filestore(
            [('empty/', None),
             ])
        listing = filestore.list_directory('empty')
        self.assertEqual([], listing)

    def test_listing_directory_of_a_file(self):
        # If a listing is attempted of a file, then None is returned.
        filestore = self.make_filestore(
            [('some-file', b'content'),
             ])
        listing = filestore.list_directory('some-file')
        self.assertIs(None, listing)

    def test_listing_directory_root(self):
        filestore = self.make_filestore(
            [('some-file', b'content'),
             ('another-file', b'a'),
             ('directory/', None),
             ('directory/subfile', b'b'),
             ])
        listing = filestore.list_directory(None)
        self.assertEqual(
            ['another-file', 'directory', 'some-file'],
            sorted(f.base_name for f in listing))

    def test_listing_directory_subdir(self):
        filestore = self.make_filestore(
            [('some-file', b'content'),
             ('another-file', b'a'),
             ('directory/', None),
             ('directory/subfile', b'b'),
             ('directory/another', b'a'),
             ])
        listing = filestore.list_directory('directory')
        self.assertEqual(
            ['another', 'subfile'],
            sorted(f.base_name for f in listing))

    def test_last_modified(self):
        # Make sure that the timestamp and author are recorded.
        start = datetime.utcnow() - timedelta(5)
        filestore = self.make_filestore()
        filestore.update_file(
            'new-file.txt',
            b'some content',
            'Test Author <test@example.com>',
            None)
        curr = filestore.get_file('new-file.txt')
        end = datetime.utcnow() + timedelta(5)
        # A new line is added to the end too.
        self.assertEqual(
            'Test Author <test@example.com>',
            curr.last_modified_by)

        self.assertTrue(
            start <= curr.last_modified_date <= end,
            '%r <= %r <= %r' % (start, curr.last_modified_date, end))



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""View classes to control the rendering of the content."""

from wikkid.interface.resource import IDirectoryResource
from wikkid.view.base import DirectoryBreadcrumbView
from wikkid.view.urls import canonical_url


class ListingItem(object):
    """An item to be shown in the directory listing."""

    def __init__(self, context, request, view, css_class, name=None):
        self.context = context
        self.url = canonical_url(self.context, request, view)
        if name is None:
            name = context.base_name
        self.name = name
        self.css_class = css_class


class DirectoryListingPage(DirectoryBreadcrumbView):
    """The directory listing shows the content in the directory.

    This view is shown if there is no matching wiki apge the same name as the
    directory (i.e. with '.txt' on the end).
    """

    for_interface = IDirectoryResource
    name = 'listing'
    is_default = False
    template = 'view_directory'

    def before_render(self):
        """Get the listing and split it into directories and files."""
        directories = []
        files = []
        for item in self.context.get_listing():
            if IDirectoryResource.providedBy(item):
                directories.append(item)
            else:
                files.append(item)

        def sort_key(item):
            return item.base_name.lower()

        items = []
        # If we are looking at / don't add a parent dir.
        if self.context.path != '/':
            parent = self.context.parent
            items.append(
                ListingItem(parent, self.request, 'listing', 'up', name='..'))
        for item in sorted(directories, key=sort_key):
            items.append(
                ListingItem(item, self.request, 'listing', 'directory'))
        for item in sorted(files, key=sort_key):
            items.append(ListingItem(item, self.request, None, 'file'))
        self.items = items

    @property
    def content(self):
        return 'Directory listing for %s' % self.path

    @property
    def title(self):
        """The title is just the directory path."""
        dir_name = self.context.get_dir_name()
        if dir_name is None:
            return 'Wiki Root'
        else:
            return dir_name



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the wikkid.formatter.textileformatter module."""

from textwrap import dedent

from bs4 import BeautifulSoup

# The textile formatter is optional.  If the MarkdownFormatter is not
# available, the tests are skipped.
try:
    from wikkid.formatter.textileformatter import TextileFormatter
    has_textile = True
except ImportError:
    has_textile = False
from wikkid.tests import TestCase


class TestTextileFormatter(TestCase):
    """Tests for the textile formatter."""

    def setUp(self):
        TestCase.setUp(self)
        if has_textile:
            self.formatter = TextileFormatter()
        else:
            self.skip('textile formatter not available')

    def test_detailed_headings(self):
        text = dedent("""\
            h1. Heading 1

            h2. Heading 2

            h3. Heading 3

            h4. Heading 4

            h5. Heading 5

            h6. Heading 6
            """)
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result)
        self.assertEqual('Heading 1', soup.h1.string)
        self.assertEqual('Heading 2', soup.h2.string)
        self.assertEqual('Heading 3', soup.h3.string)
        self.assertEqual('Heading 4', soup.h4.string)
        self.assertEqual('Heading 5', soup.h5.string)
        self.assertEqual('Heading 6', soup.h6.string)

    def test_inline_link(self):
        # A paragraph containing a wiki word.
        text = 'A link to the "FrontPage":http://127.0.0.1 helps.'
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result)
        self.assertEqual('http://127.0.0.1', soup.a['href'])

    def test_emphasis(self):
        text = 'We can have _emphasis_ and *strong* as well!'
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result)
        self.assertEqual('emphasis', soup.em.string)
        self.assertEqual('strong', soup.strong.string)

    def test_blockquote(self):
        text = dedent('''\
            Some Text

            bq. This is a block quoted paragraph
            that spans multiple lines.

            Some more text.
            ''')
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result)
        self.assertTrue(soup.blockquote is not None)

    def test_lists(self):
        text = dedent('''\
            Some Text.

            * UL 1
            * UL 2
            * UL 3

            Some More Text!

            # OL 1
            # OL 2
            # OL 3
            ''')
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result)

        self.assertTrue(soup.ul)
        ulNodes = soup.ul.findAll('li')
        for i in range(3):
            self.assertEqual('UL %d' % (i+1), ulNodes[i].string.strip())

        self.assertTrue(soup.ol)
        olNodes = soup.ol.findAll('li')
        for i in range(3):
            self.assertEqual('OL %d' % (i+1), olNodes[i].string.strip())

    def test_code_blocks(self):
        text = dedent('''\
            Some Normal Text.

            @Some Code inside pre tags@

            More Normal text.
            ''')
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result)
        self.assertEqual(soup.code.string.strip(), 'Some Code inside pre tags')

    def test_unicode(self):
        # Test the unicode support of the textile formatter.
        text = u'\N{SNOWMAN}'
        result = self.formatter.format('format', text)
        soup = BeautifulSoup(result)
        self.assertEqual(soup.p.string.strip(), text)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2010-2012 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A means of storing execution context."""

DEFAULT_FORMAT = 'rest'
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 8080


class ExecutionContext(object):
    """Store run-time execution context data.

    This is the Encapsulate Context pattern.
    """

    def __init__(self, host=None, port=None, default_format=None,
                 script_name=None):
        """Create an execution context for the application.

        :param host: The hostname that content is being served from.
        :param port: The port that is being listened on.
        :param default_format: The default wiki format for pages that
            don't specify any.
        """
        if host is None:
            host = DEFAULT_HOST
        if port is None:
            port = DEFAULT_PORT
        if default_format is None:
            default_format = DEFAULT_FORMAT
        self.host = host
        self.port = port
        self.default_format = default_format
        # TODO: make sure the script_name if set starts with a slash and
        # doesn't finish with one.
        if script_name is None:
            script_name = ''
        self.script_name = script_name.rstrip('/')



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests the edit views."""

from testtools.matchers import Equals
from webob import Request

from wikkid.tests import TestCase
from wikkid.tests.factory import FactoryTestCase
from wikkid.view.urls import canonical_url, parse_url


class TestCanonicalUrl(FactoryTestCase):
    """Test the wikkid.view.base.canonical_url."""

    def assertUrl(self, resource, url, view_name=None, base_url=None):
        request = Request.blank('/', base_url=base_url)
        self.assertThat(canonical_url(resource, request, view_name),
                        Equals(url))

    def test_root(self):
        factory = self.make_factory()
        root = factory.get_resource_at_path('/')
        self.assertUrl(root, '/')

    def test_root_listing(self):
        factory = self.make_factory()
        root = factory.get_resource_at_path('/')
        self.assertUrl(root, '/+listing', view_name='listing')

    def test_default(self):
        factory = self.make_factory([
            ('Home.txt', 'Some content'),
            ])
        root = factory.get_resource_at_path('/')
        self.assertUrl(root.default_resource, '/Home')

    def test_default_view(self):
        factory = self.make_factory([
            ('Home.txt', 'Some content'),
            ])
        root = factory.get_resource_at_path('/')
        self.assertUrl(root.default_resource, '/Home/+edit', view_name='edit')

    def test_wiki_page(self):
        factory = self.make_factory([
            ('SomeDir/SomePage.txt', 'Some content'),
            ])
        page = factory.get_resource_at_path('/SomeDir/SomePage')
        self.assertUrl(page, '/SomeDir/SomePage')

    def test_wiki_page_view(self):
        factory = self.make_factory([
            ('SomeDir/SomePage.txt', 'Some content'),
            ])
        page = factory.get_resource_at_path('/SomeDir/SomePage')
        self.assertUrl(page, '/SomeDir/SomePage/+edit', view_name='edit')

    def test_wiki_page_full_url(self):
        factory = self.make_factory([
            ('SomeDir.txt', 'Some content'),
            ])
        page = factory.get_resource_at_path('/SomeDir.txt')
        self.assertUrl(page, '/SomeDir')

    def test_wiki_page_full_url_with_view(self):
        factory = self.make_factory([
            ('SomeDir.txt', 'Some content'),
            ])
        page = factory.get_resource_at_path('/SomeDir.txt')
        self.assertUrl(page, '/SomeDir/+edit', view_name='edit')

    def test_other_file(self):
        factory = self.make_factory([
            ('simple.py', '#!/usr/bin/python'),
            ])
        page = factory.get_resource_at_path('/simple.py')
        self.assertUrl(page, '/simple.py')

    def test_other_file_view(self):
        factory = self.make_factory([
            ('simple.py', '#!/usr/bin/python'),
            ])
        page = factory.get_resource_at_path('/simple.py')
        self.assertUrl(page, '/simple.py/+edit', view_name='edit')

    def test_missing(self):
        factory = self.make_factory()
        missing = factory.get_resource_at_path('/MissingPage')
        self.assertUrl(missing, '/MissingPage')

    def test_missing_view(self):
        factory = self.make_factory()
        missing = factory.get_resource_at_path('/MissingPage')
        self.assertUrl(missing, '/MissingPage/+edit', view_name='edit')


class TestParseUrl(TestCase):
    """Tests for wikkid.view.base.parse_url."""

    def test_root(self):
        path, view = parse_url('/')
        self.assertEqual('/', path)
        self.assertIs(None, view)

    def test_root_view(self):
        path, view = parse_url('/+listing')
        self.assertEqual('/', path)
        self.assertEqual('listing', view)

    def test_path(self):
        path, view = parse_url('/some/path')
        self.assertEqual('/some/path', path)
        self.assertIs(None, view)

    def test_path_view(self):
        path, view = parse_url('/some/path/+view')
        self.assertEqual('/some/path', path)
        self.assertEqual('view', view)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The base resource class."""

import breezy.urlutils as urlutils

from wikkid.interface.resource import IRootResource


class BaseResource(object):
    """Information about a resource."""

    def __init__(self, server, path, write_filename,
                 file_resource, dir_resource):
        self.factory = server
        self.path = path
        self.write_filename = write_filename
        self.file_resource = file_resource
        self.dir_resource = dir_resource

    @property
    def preferred_path(self):
        return self.factory.get_preferred_path(self.path)

    @property
    def base_name(self):
        return urlutils.basename(self.path)

    @property
    def dir_name(self):
        return urlutils.dirname(self.path)

    @property
    def parent(self):
        if IRootResource.providedBy(self):
            return None
        return self.factory.get_resource_at_path(self.dir_name)

    @property
    def default_resource(self):
        """Any resource should be able to get to the default resource."""
        return self.factory.get_default_resource()

    @property
    def root_resource(self):
        """Any resource should be able to get to the root resource."""
        return self.factory.get_resource_at_path('/')



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""View classes to control the rendering of the content."""

from wikkid.formatter.pygmentsformatter import PygmentsFormatter
from wikkid.interface.resource import ISourceTextFile
from wikkid.view.base import DirectoryBreadcrumbView


class SourceTextPage(DirectoryBreadcrumbView):
    """Any other non-binary file is considered other text.

    Will be rendered using pygments.
    """

    for_interface = ISourceTextFile
    name = 'view'
    is_default = True
    template = 'view_page'

    @property
    def content(self):
        formatter = PygmentsFormatter()
        return formatter.format(
            self.context.base_name, self.context.get_bytes())



#
# Copyright (C) 2010-2012 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for method and classes in wikkid.app."""

import os.path

from testtools.matchers import IsInstance
from webob.request import environ_from_url

from wikkid.app import WikkidApp
from wikkid.context import ExecutionContext
from wikkid.filestore.volatile import FileStore
from wikkid.view.missing import MissingPage
from wikkid.view.root import RootPage
from wikkid.view.wiki import WikiPage
from wikkid.tests import TestCase


class TestApp(TestCase):

    def assert_not_found(self, status, headers):
        self.assertEqual("404 Not Found", status)

    def assert_ok(self, status, headers):
        self.assertEqual("200 OK", status)

    def test_traverse_above_static_not_possible_with_relative_path(self):
        """
        Traversal above the static folder, by forging a malicious request with
        a relative path for example, is not possible.
        """
        environ = environ_from_url("/static/../page.html")
        filestore = FileStore()
        app = WikkidApp(filestore)
        app(environ, self.assert_not_found)

    def test_traverse_above_static_not_possible_with_absolute_path(self):
        """
        Traversal above the static folder, by forging a malicious request
        including an absolute path for example, is not possible.
        """
        this_file = os.path.abspath(__file__)
        environ = environ_from_url("/static/" + this_file)
        filestore = FileStore()
        app = WikkidApp(filestore)
        app(environ, self.assert_not_found)

    def test_getting_static_style_css_works(self):

        environ = environ_from_url("/static/default.css")
        filestore = FileStore()
        app = WikkidApp(filestore)
        app(environ, self.assert_ok)

    def test_getting_static_style_css_works_with_script_name(self):

        environ = environ_from_url("/test/static/default.css")
        filestore = FileStore()
        context = ExecutionContext(script_name="/test")
        app = WikkidApp(filestore, execution_context=context)
        app(environ, self.assert_ok)

    def test_getting_static_css_works_script_name_multiple_segments(self):
        environ = environ_from_url("/p/project-name/wiki/static/default.css")
        filestore = FileStore()
        context = ExecutionContext(script_name="/p/project-name/wiki")
        app = WikkidApp(filestore, execution_context=context)
        app(environ, self.assert_ok)

    def test_getting_anything_outside_script_name_fails(self):
        environ = environ_from_url("/foo/bar")
        filestore = FileStore()
        context = ExecutionContext(script_name="/test")
        app = WikkidApp(filestore, execution_context=context)
        app(environ, self.assert_not_found)

    def assertUrlIsView(self, url, view_type,
                        script_name=None, store_content=None):
        environ = environ_from_url(url)
        filestore = FileStore(store_content)
        context = ExecutionContext(script_name=script_name)
        app = WikkidApp(filestore, execution_context=context)
        view = app.get_view(environ)
        self.assertThat(view, IsInstance(view_type))

    def test_home_redirect_url_matches_script_name(self):
        self.assertUrlIsView("/test", RootPage, "/test")
        self.assertUrlIsView("/test", RootPage, "/test/")
        self.assertUrlIsView("/test/", RootPage, "/test")
        self.assertUrlIsView("/test/", RootPage, "/test/")

    def test_get_view(self):
        self.assertUrlIsView("/Home", MissingPage)

    def test_get_home_view(self):
        content = [('Home.txt', 'Welcome Home.')]
        self.assertUrlIsView("/Home", WikiPage, store_content=content)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A text to html formatter using reStuctured Text."""

from docutils.core import publish_parts
from zope.interface import implementer

from wikkid.interface.formatter import ITextFormatter


@implementer(ITextFormatter)
class RestructuredTextFormatter(object):
    """Format text as HTML using restructured text."""

    def format(self, filename, text):
        """Format the text.

        I'm almost 100% positive that this method needs more args.
        """
        parts = publish_parts(text, writer_name='html')
        return parts['html_title'] + parts['body']



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A text to html formatter using pygments."""

import cgi

from pygments import highlight
from pygments.lexers import guess_lexer_for_filename
from pygments.formatters import HtmlFormatter
from pygments.util import ClassNotFound
from zope.interface import implementer

from wikkid.interface.formatter import ITextFormatter


@implementer(ITextFormatter)
class PygmentsFormatter(object):
    """Format source files as HTML using pygments."""

    def format(self, filename, text):
        """Format the text.

        We can at a later time try to guess the lexer based on the file
        content.
        """
        try:
            lexer = guess_lexer_for_filename(filename, text)
            return highlight(text, lexer, HtmlFormatter())
        except ClassNotFound:
            return "<pre>{0}</pre>".format(cgi.escape(text))



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""View classes for the wiki root."""

from webob.exc import HTTPSeeOther

from wikkid.interface.resource import IRootResource
from wikkid.view.base import BaseView


class RootPage(BaseView):
    """The default view for the root page redirects to the home page."""

    for_interface = IRootResource
    name = 'view'
    is_default = True

    def _render(self, skin):
        """Redirect to Home (or the default page)."""
        default_resource = self.context.default_resource
        location = self.canonical_url(default_resource)
        raise HTTPSeeOther(location=location)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A WSGI application for Wikkid."""

import logging
import mimetypes
import os.path
import urllib.parse
from wsgiref.util import shift_path_info

from breezy import urlutils
from webob import Request, Response
from webob.exc import HTTPException, HTTPNotFound

from wikkid.context import ExecutionContext
from wikkid.dispatcher import get_view
from wikkid.fileutils import FileIterable
from wikkid.model.factory import ResourceFactory
from wikkid.skin.loader import Skin
from wikkid.view.urls import parse_url


def serve_file(filename):
    if os.path.exists(filename):
        basename = urlutils.basename(filename)
        content_type = mimetypes.guess_type(basename)[0]

        res = Response(content_type=content_type, conditional_response=True)
        res.app_iter = FileIterable(filename)
        res.content_length = os.path.getsize(filename)
        res.last_modified = os.path.getmtime(filename)
        # Todo: is this the best value for the etag?
        # perhaps md5 would be a better alternative
        res.etag = '%s-%s-%s' % (
            os.path.getmtime(filename), os.path.getsize(filename),
            hash(filename))
        return res

    else:
        return HTTPNotFound()


class WikkidApp(object):
    """The main wikkid application."""

    def __init__(self, filestore, skin_name=None, execution_context=None):
        if execution_context is None:
            execution_context = ExecutionContext()
        self.execution_context = execution_context
        self.filestore = filestore
        self.resource_factory = ResourceFactory(self.filestore)
        # Need to load the initial templates for the skin.
        if skin_name is None:
            skin_name = 'default'
        self.skin = Skin(skin_name)
        self.logger = logging.getLogger('wikkid')

    def preprocess_environ(self, environ):
        request = Request(environ)
        path = urllib.parse.unquote(request.path)
        script_name = self.execution_context.script_name
        # Firstly check to see if the path is the same as the script_name
        if path != script_name and not path.startswith(script_name + '/'):
            raise HTTPNotFound()

        shifted_prefix = ''
        while shifted_prefix != script_name:
            shifted = shift_path_info(environ)
            shifted_prefix = '{0}/{1}'.format(shifted_prefix, shifted)
        # Now we are just interested in the path_info having ignored the
        # script name.
        path = urllib.parse.unquote(request.path_info)
        if path == '':
            path = '/'  # Explicitly be the root (we need the /)
        return request, path

    def _get_view(self, request, path):
        """Get the view for the path specified."""
        resource_path, action = parse_url(path)
        model = self.resource_factory.get_resource_at_path(resource_path)
        return get_view(model, action, request, self.execution_context)

    def process_call(self, environ):
        """The actual implementation of dealing with the call."""
        # TODO: reject requests that aren't GET or POST
        try:
            request, path = self.preprocess_environ(environ)
        except HTTPException as e:
            return e

        if path == '/favicon.ico':
            if self.skin.favicon is not None:
                return serve_file(self.skin.favicon)
            else:
                return HTTPNotFound()

        if path.startswith('/static/'):
            if self.skin.static_dir is not None:
                static_dir = self.skin.static_dir.rstrip(os.sep) + os.sep
                static_file = os.path.abspath(
                    urlutils.joinpath(static_dir, path[8:]))
                if static_file.startswith(static_dir):
                    return serve_file(static_file)
                else:
                    return HTTPNotFound()
            else:
                return HTTPNotFound()

        try:
            view = self._get_view(request, path)
            return view.render(self.skin)
        except HTTPException as e:
            return e

    def get_view(self, environ):
        """Allow an app user to get the wikkid view for a particular call."""
        request, path = self.preprocess_environ(environ)
        return self._get_view(request, path)

    def __call__(self, environ, start_response):
        """The WSGI bit."""
        response = self.process_call(environ)
        return response(environ, start_response)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""View classes to control the rendering of wiki pages."""

from webob.exc import HTTPTemporaryRedirect

from wikkid.formatter.registry import get_wiki_formatter
from wikkid.interface.resource import IWikiTextFile
from wikkid.view.base import BaseView


def format_content(bytes, base_name, default_format):
    """ Format the content with the right formatter.

    Check the first line of the content to see if it specifies a
    formatter. The default is currently ReST, but we should have it
    configurable shortly.
    """
    content, formatter = get_wiki_formatter(bytes, default_format)
    return formatter.format(base_name, content)


class WikiPage(BaseView):
    """A wiki page is a page that is going to be rendered for viewing."""

    for_interface = IWikiTextFile
    name = 'view'
    is_default = True
    template = 'view_page'

    @property
    def content(self):
        bytes = self.context.get_bytes()
        try:
            text = bytes.decode('utf-8')
        except UnicodeDecodeError:
            try:
                text = bytes.decode('latin-1')
            except UnicodeDecodeError:
                text = bytes.decode('ascii', 'replace')

        default_format = self.execution_context.default_format
        return format_content(text, self.context.base_name, default_format)

    def _render(self, skin):
        """If the page is not being viewed with the preferred path, redirect.

        For example, /FrontPage.txt will redirect to /FrontPage
        """
        preferred = self.context.preferred_path
        if self.context.path != preferred:
            location = self.canonical_url(self.context)
            raise HTTPTemporaryRedirect(location=location)
        else:
            return super(WikiPage, self)._render(skin)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The wikkid tests for the wikkid.view package."""

import unittest


def test_suite():
    names = [
        'breadcrumbs',
        'utils',
        'edit',
        'root',
        'urls',
        'view',
        ]
    module_names = ['wikkid.tests.views.test_' + name for name in names]
    loader = unittest.TestLoader()
    return loader.loadTestsFromNames(module_names)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the wikkid.view module."""

from testtools import TestCase

from wikkid.view.utils import expand_wiki_name, title_for_filename


class TestExpandWikiName(TestCase):
    """Tests for expand_wiki_name."""

    def test_expand_wiki_name(self):
        self.assertEqual('simple.txt', expand_wiki_name('simple.txt'))
        self.assertEqual('nonMatching', expand_wiki_name('nonMatching'))
        self.assertEqual('README', expand_wiki_name('README'))
        self.assertEqual('Home', expand_wiki_name('Home'))
        self.assertEqual('Front Page', expand_wiki_name('FrontPage'))
        self.assertEqual('FrontPage.txt', expand_wiki_name('FrontPage.txt'))
        self.assertEqual('A Simple Page', expand_wiki_name('ASimplePage'))
        self.assertEqual('FTP Example', expand_wiki_name('FTPExample'))
        self.assertEqual(
            'A Simple FTP Example',
            expand_wiki_name('ASimpleFTPExample'))


class TesttitleForFilename(TestCase):
    """Tests for title_for_filename."""

    def test_title_for_filename(self):
        self.assertEqual('simple', title_for_filename('simple.txt'))
        self.assertEqual('simple.cpp', title_for_filename('simple.cpp'))
        self.assertEqual('Front Page', title_for_filename('FrontPage'))
        self.assertEqual('Front Page', title_for_filename('FrontPage.txt'))
        self.assertEqual('FrontPage.cpp', title_for_filename('FrontPage.cpp'))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A text to html formatter using markdown."""

import markdown
from zope.interface import implementer

from wikkid.interface.formatter import ITextFormatter


@implementer(ITextFormatter)
class MarkdownFormatter(object):
    """Format source files as HTML using markdown."""

    def format(self, filename, text):
        """Format the text.
        """
        md = markdown.Markdown(safe_mode='replace')
        return md.convert(text)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The skin is the outer look and feel of the wikkid wiki pages.

It is intended that the user will be able to use one of the pre-defined skins
(of which there is only the default right now) or provide a directory to their
own skin as a command line argument.
"""

import logging
import os.path

import breezy.urlutils as urlutils
from jinja2 import Environment, PackageLoader


class Skin(object):
    """A Wikkid wiki skin."""

    def __init__(self, skin_name=None):
        """Load the required templates."""
        # Need to load the initial templates for the skin.
        if skin_name is None:
            skin_name = 'default'
        self.logger = logging.getLogger('wikkid')
        # TODO: if we are using a user defined directory for the skin, here is
        # where we'd use a different loader.
        loader = PackageLoader('wikkid.skin', skin_name)
        self.env = Environment(loader=loader)
        self.templates = {
            'view_page': self.env.get_template('page.html'),
            'edit_page': self.env.get_template('edit.html'),
            'view_directory': self.env.get_template('directory_listing.html'),
            'missing': self.env.get_template('missing-page.html'),
            'missing-dir': self.env.get_template('missing-directory.html')
            }
        module_location = urlutils.dirname(__file__)
        self.dir_name = urlutils.joinpath(module_location, skin_name)

    def get_template(self, template_name):
        return self.templates[template_name]

    @property
    def favicon(self):
        location = os.path.abspath(
            urlutils.joinpath(self.dir_name, 'favicon.ico'))
        if os.path.exists(location):
            return location
        else:
            return None

    @property
    def static_dir(self):
        location = os.path.abspath(
            urlutils.joinpath(self.dir_name, 'static'))
        if os.path.exists(location):
            return location
        else:
            return None



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the wikkid bzr user."""

from breezy.tests import TestCaseWithTransport
from webob import Request, Response

from wikkid.interface.user import IUser
from wikkid.tests import ProvidesMixin
from wikkid.user.bzr import LocalBazaarUserMiddleware


class TestLocalUserMiddleware(TestCaseWithTransport, ProvidesMixin):

    def setUp(self):
        super(TestLocalUserMiddleware, self).setUp()
        self.user = None

    def app_func(self, environment, start_response):
        self.user = environment['wikkid.user']
        return Response('done')(environment, start_response)

    def test_user_is_set(self):
        branch = self.make_branch_and_tree('.').branch
        req = Request.blank('/')
        app = LocalBazaarUserMiddleware(self.app_func, branch)
        req.get_response(app)
        self.assertIsNot(None, self.user)
        self.assertProvides(self.user, IUser)

    def test_user_attributes(self):
        branch = self.make_branch_and_tree('.').branch
        branch.get_config().set_user_option(
            'email', 'Test User <test@example.com>')
        req = Request.blank('/')
        app = LocalBazaarUserMiddleware(self.app_func, branch)
        req.get_response(app)
        self.assertEqual(
            'Test User <test@example.com>', self.user.committer_id)
        self.assertEqual('Test User', self.user.display_name)
        self.assertEqual('test@example.com', self.user.email)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the model objects."""

from operator import attrgetter

from wikkid.interface.resource import (
    IDefaultPage, IDirectoryResource, IMissingResource, IRootResource,
    IWikiTextFile)
from wikkid.tests.factory import FactoryTestCase


class TestBaseRootResource(FactoryTestCase):
    """Tests for BaseResource.root_resource."""

    def test_root_can_get_root(self):
        """Even the root resource can get the root resource."""
        factory = self.make_factory()
        resource = factory.get_resource_at_path('/')
        root = resource.root_resource
        self.assertProvides(root, IRootResource)

    def test_missing_can_get_root(self):
        """Even the root resource can get the root resource."""
        factory = self.make_factory()
        resource = factory.get_resource_at_path('/MissingPage')
        root = resource.root_resource
        self.assertProvides(root, IRootResource)


class TestBaseDefaultResource(FactoryTestCase):
    """Tests for BaseResource.default_resource."""

    def test_default_when_missing(self):
        """If the default is missing, then a missing resource is returned for
        the default path.
        """
        factory = self.make_factory()
        resource = factory.get_resource_at_path('/')
        home = resource.default_resource
        self.assertProvides(home, IMissingResource)
        self.assertProvides(home, IDefaultPage)
        self.assertEqual('/Home', home.preferred_path)

    def test_default_when_exists(self):
        """If the default exists, the default wiki page is returned."""
        factory = self.make_factory([
                ('Home.txt', 'Some content'),
                ])
        resource = factory.get_resource_at_path('/')
        home = resource.default_resource
        self.assertProvides(home, IWikiTextFile)
        self.assertProvides(home, IDefaultPage)
        self.assertEqual('/Home', home.preferred_path)

    def test_default_different_name(self):
        """If the default name has been overridden, then the default_resource
        attribute returns the correct name.
        """
        factory = self.make_factory([
                ('FrontPage.txt', 'Some content'),
                ])
        factory.DEFAULT_PATH = 'FrontPage'
        resource = factory.get_resource_at_path('/')
        home = resource.default_resource
        self.assertProvides(home, IWikiTextFile)
        self.assertProvides(home, IDefaultPage)
        self.assertEqual('/FrontPage', home.preferred_path)


class TestBaseParent(FactoryTestCase):
    """Tests for the BaseResource.parent attribute."""

    def test_parent_for_root(self):
        """Root shouldn't have a parent."""
        factory = self.make_factory()
        root = factory.get_root_resource()
        self.assertIs(None, root.parent)

    def test_parent_for_default(self):
        """The default's parent should be root."""
        factory = self.make_factory()
        home = factory.get_default_resource()
        parent = home.parent
        self.assertProvides(parent, IRootResource)

    def test_parent_for_page(self):
        """Any page in the root directory has root as its parent."""
        factory = self.make_factory([
                ('SomeDir.txt', 'Some content'),
                ])
        home = factory.get_resource_at_path('/SomeDir')
        parent = home.parent
        self.assertProvides(parent, IRootResource)

    def test_parent_for_subpage(self):
        """The path of the parent should be the parent directory."""
        factory = self.make_factory([
                ('SomeDir/', None),
                ('SomeDir.txt', 'Some content'),
                ])
        missing = factory.get_resource_at_path('/SomeDir/NoPage')
        parent = missing.parent
        expected = factory.get_resource_at_path('/SomeDir')
        self.assertEquals(expected.preferred_path, parent.preferred_path)


class TestDirectoryResource(FactoryTestCase):

    def test_implements_interface(self):
        """DirectoryResource implements IDirectoryResource."""
        factory = self.make_factory([
                ('SomeDir/', None),
                ])
        dir_resource = factory.get_resource_at_path('/SomeDir')
        self.assertProvides(dir_resource, IDirectoryResource)

    def test_directory_and_pages(self):
        """Both the directory and wiki page are returned."""
        factory = self.make_factory([
                ('SomeDir/', None),
                ('SomeDir.txt', 'Some content'),
                ])
        dir_resource = factory.get_resource_at_path('/')
        listing = dir_resource.get_listing()
        some_dir, some_wiki = sorted(
            listing, key=attrgetter('path'))
        self.assertEqual('/SomeDir', some_dir.path)
        self.assertEqual('/SomeDir.txt', some_wiki.path)

    def test_user_for_file(self):
        """Test that the user is a user object."""
        factory = self.make_factory()
        page = factory.get_resource_at_path('/testing')
        page.put_bytes(
            'hello world', 'Test User <test@example.com>', None, None)
        new_page = factory.get_resource_at_path('/testing')
        user = new_page.last_modified_by
        self.assertEqual('Test User', user.display_name)
        self.assertEqual('test@example.com', user.email)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The source text class.

A source text file is a text file that isn't a wiki file.
"""

from zope.interface import implementer

from wikkid.model.textfile import TextFile
from wikkid.interface.resource import ISourceTextFile


@implementer(ISourceTextFile)
class SourceTextFile(TextFile):
    """A text file that isn't a wiki page."""



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Wikkid filestores.

The volatile filestore is primarily used for tests, and the bzr filestore is
where all the main work happens.
"""


class FileExists(Exception):
    """A file was found where a directory is wanted."""


class UpdateConflicts(Exception):
    """Conflicts were found during updating."""
    def __init__(self, content, basis_rev):
        Exception.__init__(self)
        self.content = content
        self.basis_rev = basis_rev



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The missing resource class.

A missing resource is a file that has been asked for that doesn't exist.

A model object exists for a missing resource as in a wiki you can have views
on things that aren't there, like a page asking if you want to make a wiki
page there.
"""

from zope.interface import implementer

from wikkid.model.file import UpdatableResource
from wikkid.interface.resource import IMissingResource


@implementer(IMissingResource)
class MissingResource(UpdatableResource):
    """Information about a resource."""

    def __repr__(self):
        return "<MissingResource '%s'>" % self.path



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the wikkid.formatter.rest module."""

from textwrap import dedent
from bs4 import BeautifulSoup

from wikkid.formatter.restformatter import RestructuredTextFormatter
from wikkid.tests import TestCase


class TestRestructuredTextFormatter(TestCase):
    """Tests for the ReST formatter."""

    def setUp(self):
        TestCase.setUp(self)
        self.formatter = RestructuredTextFormatter()

    def test_simple_text(self):
        # A simple heading and a paragraph.
        text = dedent("""\
            Nice Heading
            ============

            Simple sentence.
            """)
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result, features="lxml")
        self.assertEqual('Nice Heading', soup.h1.string)
        self.assertEqual('Simple sentence.', soup.p.string)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Interface for the text formatters."""


from zope.interface import Interface


class ITextFormatter(Interface):
    """A text formatter takes plain text and makes HTML of some form."""

    def format(filename, text):
        """Takes text, and returns HTML."""



#
# Copyright (C) 2012 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the wikkid.filestore.git.FileStore."""

from dulwich.repo import MemoryRepo

from wikkid.filestore.git import (
    FileStore,
    )
from wikkid.tests import ProvidesMixin, TestCase
from wikkid.tests.filestore import TestFileStore


class TestGitFileStore(TestCase, ProvidesMixin, TestFileStore):
    """Tests for the git filestore and files."""

    def make_filestore(self, contents=None):
        repo = MemoryRepo()
        fs = FileStore(repo)
        if contents:
            for (path, contents) in contents:
                if contents is None:
                    # Directory
                    continue
                fs.update_file(
                    path, contents,
                    author="Somebody <test@example.com>",
                    parent_revision=None,
                    commit_message="Added by make_filestore")
        return fs

    def test_empty(self):
        # Empty files do not have line endings, but they can be saved
        # nonetheless.
        filestore = self.make_filestore(
            [('test.txt', b'several\nlines\nof\ncontent')])
        f = filestore.get_file('test.txt')
        base_rev = f.last_modified_in_revision
        filestore.update_file(
            'test.txt', b'', 'Test Author <test@example.com>', base_rev)
        curr = filestore.get_file('test.txt')
        self.assertEqual(b'', curr.get_content())

    def test_listing_directory_empty(self):
        filestore = self.make_filestore(
            [('empty/', None)])
        listing = filestore.list_directory('empty')
        self.assertIs(None, listing)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Interfaces for the different resource types."""

from zope.interface import Attribute, Interface


class IDefaultPage(Interface):
    """A marker interface for the default wiki page."""


class IResource(Interface):
    """The base resource interface."""

    path = Attribute('The full path for the resource.')

    preferred_path = Attribute('The preferred path for the resource.')

    write_filename = Attribute(
        'The full path of the file to write to in the filestore. '
        'This is either the filename as it directly corresponds to the '
        'path, or a related wiki page location.')

    root_resource = Attribute(
        'The root resource is the object that represents the root of the wiki.'
        )

    default_resource = Attribute(
        'The default resource is the default wiki page.'
        )


class IUpdatableResource(IResource):
    """Reflects either a file either missing or actual."""

    def put_bytes(bytes, committer, rev_id, commit_msg):
        """Update the content of the resource with the bytes specified.

        :param bytes: A byte string reflecting the new file content.
        :param committer: The committer string that will be used.
        :param rev_id: The base revision id for the text being edited.
            None when adding a new file.
        :param commit_msg: The message to associate with this edit.
        """


class IFileResource(IUpdatableResource):
    """A resource that relates to a file in the filestore."""

    # TODO: think of a better variable name.
    file_resource = Attribute(
        'An IFile representing the file in the filestore.')

    mimetype = Attribute('The mimetype of the file.')

    def get_bytes():
        """Returns the bytes of the binary file."""

    last_modified_in_revision = Attribute(
        'The revision id where the file was last modified.')

    last_modified_by = Attribute(
        'The author of the revision that last modified the file.')

    last_modified_date = Attribute(
        'The timestamp of the revision that last modified the file. '
        'This is a naive datetime object in UTC.')


class IDirectoryResource(IResource):
    """A resource that relates to a file in the filestore."""

    def get_dir_name():
        """Get the full directory name.

        This is the full path for the directory from the root.
        """

    # TODO: think of a better variable name.
    dir_resource = Attribute(
        'An IFile representing the directory in the filestore.')

    def get_listing():
        """Returns a list of objects."""


class IRootResource(IDirectoryResource):
    """A special resource relating to the root object in the wiki."""

    has_home_page = Attribute(
        'True if there is a home wiki page defined.')


class IBinaryFile(IFileResource):
    """A marker interface for binary files."""


class ITextFile(IFileResource):
    """A marker interface for text files."""


class IWikiTextFile(ITextFile):
    """A marker interface for a wiki text file."""


class ISourceTextFile(ITextFile):
    """A marker interface for a non-wiki text file."""


class IMissingResource(IResource):
    """A resource that doesn't exist in the filestore."""



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Interfaces relating to users of the wiki."""

from zope.interface import Attribute, Interface


class IUserFactory(Interface):
    """A factory to create `IUser`s."""

    def create(request):
        """Returns an `IUser`."""


class IUser(Interface):
    """Information about the editing user.

    Note: probably implementations
     - test identity
     - bzr identity
     - anonymous identity
     - launchpad identity
     - session identity
    """
    email = Attribute("The user's email adderss.")
    display_name = Attribute(
        "The name that is shown through the user interface.")
    committer_id = Attribute("The user's name and email address.")



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for wikkid.formatter.registry."""

from textwrap import dedent

from wikkid.formatter.registry import get_wiki_formatter
from wikkid.tests import TestCase


class TestGetWikiFormtter(TestCase):
    """Tests for get_wiki_formatter."""

    def test_default_missing(self):
        self.assertRaises(
            KeyError,
            get_wiki_formatter,
            'some content',
            'missing')

    def test_default_exists(self):
        content, formatter = get_wiki_formatter(
            'some content', 'rest')
        self.assertEqual('some content', content)
        self.assertEqual(
            'RestructuredTextFormatter', formatter.__class__.__name__)

    def test_specify_missing(self):
        content, formatter = get_wiki_formatter(
            dedent("""\
                # missing
                some content
                """), 'rest')
        self.assertEqual('# missing\nsome content\n', content)
        self.assertEqual(
            'RestructuredTextFormatter', formatter.__class__.__name__)

    def test_specify_extra_whitespace(self):
        content, formatter = get_wiki_formatter(
            dedent("""\
                #\t\tpygments
                some content
                """), 'rest')
        self.assertEqual('some content\n', content)
        self.assertEqual('PygmentsFormatter', formatter.__class__.__name__)

    def test_specify_extra_params(self):
        content, formatter = get_wiki_formatter(
            dedent("""\
                # pygments extra params
                some content
                """), 'rest')
        self.assertEqual('some content\n', content)
        self.assertEqual('PygmentsFormatter', formatter.__class__.__name__)

    def test_specify_pygments_case_insensitive(self):
        content, formatter = get_wiki_formatter(
            dedent("""\
                # pygments
                some content
                """), 'rest')
        self.assertEqual('some content\n', content)
        self.assertEqual('PygmentsFormatter', formatter.__class__.__name__)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The wikkid tests for the wikkid.formatter package."""

import unittest


def test_suite():
    names = [
        'registry',
        'rest',
        'markdown',
        'textile',
        ]
    module_names = ['wikkid.tests.formatters.test_' + name for name in names]
    loader = unittest.TestLoader()
    suite = loader.loadTestsFromNames(module_names)

    return suite



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The wikkid tests and test only code."""

from zope.interface import implementer

from wikkid.interface.user import IUser


class TestUserFactory(object):
    """Right now, user factories don't do anything."""


@implementer(IUser)
class TestUser(object):
    """A test user that implements the interface."""

    def __init__(self, email, display_name):
        self.email = email
        self.display_name = display_name
        self.committer_id = "{0} <{1}>".format(email, display_name)


class TestRequest(object):
    """A fake request object."""



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests the display views."""

from wikkid.tests.factory import ViewTestCase


class TestView(ViewTestCase):
    """Test the display view."""

    def test_wiki_last_modified_by(self):
        """Test that the last committer is displayed properly"""
        factory = self.make_factory([
                ('SomePage/SubPage/Nested.txt', 'some text')])
        view = self.get_view(factory, '/SomePage/SubPage/Nested', 'view')
        user = view.last_modified_by
        self.assertEqual('First User', user.display_name)

    def test_other_last_modified_by(self):
        """Test that the last committer is displayed properly"""
        factory = self.make_factory([
                ('test.py', 'some text')])
        view = self.get_view(factory, '/test.py')
        user = view.last_modified_by
        self.assertEqual('First User', user.display_name)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Base classes for other filestores to use."""

import mimetypes

import breezy.urlutils as urlutils

from wikkid.interface.filestore import FileType


class BaseFile(object):
    """Provide common fields and methods and properties for files."""

    def __init__(self, path):
        self.path = path
        self.base_name = urlutils.basename(path)
        self._mimetype = mimetypes.guess_type(self.base_name)[0]

    @property
    def mimetype(self):
        """If the file_type is a directory, return None."""
        if self.file_type == FileType.DIRECTORY:
            return None
        else:
            return self._mimetype



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The server class for the wiki."""

import logging

from breezy.urlutils import basename, dirname, joinpath
from zope.interface import directlyProvides

from wikkid.interface.filestore import FileType
from wikkid.interface.resource import IDefaultPage
from wikkid.model.binary import BinaryResource
from wikkid.model.directory import DirectoryResource
from wikkid.model.missing import MissingResource
from wikkid.model.root import RootResource
from wikkid.model.sourcetext import SourceTextFile
from wikkid.model.wikitext import WikiTextFile


class ResourceFactory(object):
    """Factory to create the model objects used by the views."""

    DEFAULT_PATH = 'Home'

    def __init__(self, filestore):
        """Construct the factory.

        :param filestore: An `IFileStore` instance.
        :param user_factory: A factory to create users.
        :param skin_name: The name of a skin to use.
        """
        self.filestore = filestore
        self.logger = logging.getLogger('wikkid')

    def get_resource(self, path, file_path, file_resource, dir_resource):
        """Return the correct type of resource based on the params."""
        filename = basename(file_path)
        if path == '/':
            return RootResource(
                self, path, file_path, file_resource, None)
        elif file_resource is not None:
            # We are pointing at a file.
            file_type = file_resource.file_type
            if file_type == FileType.BINARY_FILE:
                # Binary resources have no associated directory.
                return BinaryResource(
                    self, path, file_path, file_resource, None)
            # This is known to be not entirely right.
            if filename.endswith('.txt') or '.' not in file_resource.base_name:
                return WikiTextFile(
                    self, path, file_path, file_resource,
                    dir_resource)
            else:
                return SourceTextFile(
                    self, path, file_path, file_resource, None)
        elif dir_resource is not None:
            return DirectoryResource(
                self, path, file_path, None, dir_resource)
        else:
            return MissingResource(
                self, path, file_path, None, None)

    def get_default_resource(self):
        """Return the Home resource."""
        return self.get_resource_at_path('/' + self.DEFAULT_PATH)

    def get_root_resource(self):
        """Return the root resource."""
        return self.get_resource_at_path('/')

    def get_resource_at_path(self, path):
        """Get the resource from the filestore for the specified path.

        The path starts with a slash as proveded through the url traversal,
        the filestore does not expect nor want a leading slash.  It is the
        responsibility of this method to remove the leading slash.
        """
        assert path.startswith('/')
        file_path = path[1:]
        is_default = False
        if file_path == '' or file_path == self.DEFAULT_PATH:
            file_path = self.DEFAULT_PATH
            is_default = True

        dir_resource = None
        file_resource = self.filestore.get_file(file_path)
        # If the resource exists and is a file, we are done.
        if file_resource is not None:
            if file_resource.file_type != FileType.DIRECTORY:
                return self.get_resource(path, file_path, file_resource, None)
            else:
                dir_resource = file_resource
                file_resource = None

        if '.' not in basename(file_path):
            file_path += '.txt'
            file_resource = self.filestore.get_file(file_path)

        resource = self.get_resource(
            path, file_path, file_resource, dir_resource)
        if is_default:
            directlyProvides(resource, IDefaultPage)
        return resource

    def get_preferred_path(self, path):
        """Get the preferred path for the path passed in.

        If the path ends with '.txt' and doesn't have any other '.'s in the
        basename, then we prefer to access that file without the '.txt'.

        If the resulting path is the default path, then the preferred path
        should be '/Home' providing Home is the default path..
        """
        filename = basename(path)
        if filename.endswith('.txt'):
            filename = filename[:-4]

        if path == '/':
            return '/' + self.DEFAULT_PATH
        elif '.' in filename:
            return path
        else:
            return joinpath(dirname(path), filename)



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""View URL functions."""

import re


VIEW_MATCHER = re.compile(r'^(.*)/\+(\w+)$')


def parse_url(path):
    """Convert a path into a resource path and a view."""
    match = VIEW_MATCHER.match(path)
    if match is not None:
        resource_path, view = match.groups()
        if resource_path == '':
            resource_path = '/'
        return (resource_path, view)
    else:
        return (path, None)


def canonical_url(context, request, view=None):
    """The one true URL for the context object."""
    path = context.preferred_path
    if view is None:
        return '{0}{1}'.format(request.script_name, path)
    else:
        if path == '/':
            path = ''
        return '{0}{1}/+{2}'.format(request.script_name, path, view)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The wiki text class.

A text file that contains text that will be formatted into HTML using one of
the formatters.
"""

from zope.interface import directlyProvides, implementer

from wikkid.model.directory import DirectoryMethods
from wikkid.model.textfile import TextFile
from wikkid.interface.resource import IDirectoryResource, IWikiTextFile


@implementer(IWikiTextFile)
class WikiTextFile(TextFile, DirectoryMethods):
    """A text file that represents a wiki page."""

    def __init__(self, server, path, write_filename,
                 file_resource, dir_resource):
        super(WikiTextFile, self).__init__(
            server, path, write_filename, file_resource, dir_resource)
        if dir_resource is not None:
            directlyProvides(self, IDirectoryResource)

    def __repr__(self):
        return "<WikiTextFile '%s'>" % self.path



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The base view class."""

import logging

from webob import Response

from wikkid.dispatcher import register_view
from wikkid.view.urls import canonical_url
from wikkid.view.utils import title_for_filename
from wikkid.interface.resource import IDefaultPage, IRootResource


class BaseViewMetaClass(type):
    """This metaclass registers the view with the view registry."""

    def __new__(cls, classname, bases, classdict):
        """Called when defining a new class."""
        instance = type.__new__(cls, classname, bases, classdict)
        register_view(instance)
        return instance


class Breadcrumb(object):
    """Breadcrumbs exist to give the user quick links up the path chain."""

    def __init__(self, context, request, view=None, title=None):
        self.path = canonical_url(context, request, view)
        if title is None:
            self.title = title_for_filename(context.base_name)
        else:
            self.title = title


class BaseView(object, metaclass=BaseViewMetaClass):
    """The base view class.

    This is an abstract base class.
    """

    def __init__(self, context, request, execution_context):
        self.execution_context = execution_context
        self.context = context
        self.request = request
        if request is not None:
            self.user = request.environ.get('wikkid.user', None)
        self.logger = logging.getLogger('wikkid')

    def _create_breadcrumbs(self):
        crumbs = [Breadcrumb(self.context, self.request)]
        current = self.context.parent
        while not IRootResource.providedBy(current):
            crumbs.append(Breadcrumb(current, self.request))
            current = current.parent
        # And add in the default page if the context isn't the default.
        if not IDefaultPage.providedBy(self.context):
            crumbs.append(Breadcrumb(current.default_resource, self.request))
        return reversed(crumbs)

    def initialize(self):
        """Provide post-construction initialization."""

    @property
    def breadcrumbs(self):
        return self._create_breadcrumbs()

    @property
    def title(self):
        return title_for_filename(self.context.base_name)

    @property
    def last_modified_by(self):
        return self.context.last_modified_by

    @property
    def last_modified_date(self):
        last_modified = self.context.last_modified_date
        if last_modified is None:
            return None
        return last_modified.strftime('%Y-%m-%d %H:%M:%S')

    def before_render(self):
        """A hook to do things before rendering."""

    def canonical_url(self, context, view=None):
        return canonical_url(context, self.request, view)

    def template_args(self):
        """Needs to be implemented in the derived classes.

        :returns: A dict of values.
        """
        return {
            'view': self,
            'user': self.user,
            'context': self.context,
            'request': self.request,
            'canonical_url': self.canonical_url,
            }

    def _render(self, skin):
        """Get the template and render with the args.

        If a template isn't going to be used or provide the conent,
        this is the method to override.
        """
        template = skin.get_template(self.template)
        content = template.render(**self.template_args())
        # Return the encoded content.
        return self.make_response(content.encode('utf-8'))

    def make_response(self, body):
        """Construct the response object for this request.

        :param body: The body of the response, as a unicode string.
        :return: A `Response` object.
        """
        return Response(body)

    def render(self, skin):
        """Render the page.

        Return a tuple of content type and content.
        """
        self.before_render()
        return self._render(skin)


class DirectoryBreadcrumbView(BaseView):
    """A view that uses the directories as breadcrumbs."""

    def _create_breadcrumbs(self):
        crumbs = []
        current = self.context
        view = None
        while not IRootResource.providedBy(current):
            crumbs.append(Breadcrumb(
                    current, self.request, view, title=current.base_name))
            current = current.parent
            # Add listings to subsequent urls.
            view = 'listing'
        # Add in the root dir.
        crumbs.append(Breadcrumb(current, self.request, 'listing',
                                 title='wiki root'))
        # And add in the default page.
        crumbs.append(Breadcrumb(current.default_resource, self.request))
        return reversed(crumbs)



#
# Copyright (C) 2010-2012 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for method and classes in wikkid.context."""

from testtools.matchers import Equals

from wikkid.context import ExecutionContext
from wikkid.tests import TestCase


class TestContext(TestCase):

    def test_empty_script_name(self):
        context = ExecutionContext()
        self.assertThat(context.script_name, Equals(''))

    def test_script_name(self):
        context = ExecutionContext(script_name='/foo')
        self.assertThat(context.script_name, Equals('/foo'))

    def test_script_name_strips_trailing_slash(self):
        context = ExecutionContext(script_name='/foo/')
        self.assertThat(context.script_name, Equals('/foo'))



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Test views for the root object."""

from bs4 import BeautifulSoup
from testtools.matchers import Equals
from webob.exc import HTTPSeeOther

from wikkid.skin.loader import Skin
from wikkid.tests.factory import ViewTestCase


class TestRootViews(ViewTestCase):
    """Test the views on the root object."""

    def test_root_redirects(self):
        """Going to / redirects to the Home page."""
        factory = self.make_factory()
        view = self.get_view(factory, '/')
        error = self.assertRaises(
            HTTPSeeOther,
            view.render,
            None)
        self.assertEqual('/Home', error.headers['Location'])

    def test_root_redirects_with_script_name(self):
        """Redirection works and respects the script name"""
        factory = self.make_factory()
        view = self.get_view(factory, '/', base_url='/p/test')
        error = self.assertRaises(
            HTTPSeeOther,
            view.render,
            None)
        self.assertEqual('/p/test/Home', error.headers['Location'])

    def test_home_rendering(self):
        """Render the home page and test the elements."""
        factory = self.make_factory()
        view = self.get_view(factory, '/Home')
        content = view.render(Skin('default'))
        soup = BeautifulSoup(content.text, features="lxml")
        [style] = soup.find_all('link', {'rel': 'stylesheet'})
        self.assertThat(style['href'], Equals('/static/default.css'))

    def test_home_rendering_with_script_name(self):
        """Render the home page and test the elements."""
        factory = self.make_factory()
        view = self.get_view(factory, '/Home', base_url='/p/test')
        content = view.render(Skin('default'))
        soup = BeautifulSoup(content.text, features="lxml")
        [style] = soup.find_all('link', {'rel': 'stylesheet'})
        self.assertThat(style['href'], Equals('/p/test/static/default.css'))



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A method of iterating over file contents

This stops us reading the entire file into memory when we serve it.
"""


class FileIterable(object):
    """An iterable file-like object."""
    def __init__(self, filename, start=None, stop=None):
        self.filename = filename
        self.start = start
        self.stop = stop

    def __iter__(self):
        return FileIterator(self.filename, self.start, self.stop)

    def app_iter_range(self, start, stop):
        return self.__class__(self.filename, start, stop)


class FileIterator(object):
    """Iterate over a file.

    FileIterable provides a simple file iterator, optionally allowing the
    user to specify start and end ranges for the file.
    """

    chunk_size = 4096

    def __init__(self, filename, start, stop):
        self.filename = filename
        self.fileobj = open(self.filename, 'rb')
        if start:
            self.fileobj.seek(start)
        if stop is not None:
            self.length = stop - start
        else:
            self.length = None

    def __iter__(self):
        return self

    def __next__(self):
        if self.length is not None and self.length <= 0:
            raise StopIteration
        chunk = self.fileobj.read(self.chunk_size)
        if not chunk:
            raise StopIteration
        if self.length is not None:
            self.length -= len(chunk)
            if self.length < 0:
                # Chop off the extra:
                chunk = chunk[:self.length]
        return chunk



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Views associated with binary files."""

from webob import Response

from wikkid.interface.resource import IBinaryFile
from wikkid.view.base import BaseView


class BinaryFile(BaseView):
    """Renders a binary file with its mimetype."""

    for_interface = IBinaryFile
    name = 'view'
    is_default = True

    def _render(self, skin):
        return Response(
            body=self.context.get_bytes(),
            content_type=self.context.mimetype)



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A user factory and user class which uses the bzr identity from the
local bazaar config."""

import email
import logging

from webob import Request
from zope.interface import implementer

from wikkid.interface.user import IUser, IUserFactory
from wikkid.user.baseuser import BaseUser


def create_bzr_user_from_author_string(author):
    name, address = email.utils.parseaddr(author)
    if name:
        display_name = name
    else:
        display_name = address
    return User(address, display_name, author)


class LocalBazaarUserMiddleware(object):
    """A middleware to inject a user into the environment."""

    def __init__(self, app, branch):
        self.app = app
        config = branch.get_config()
        self.user = create_bzr_user_from_author_string(config.username())

    def __call__(self, environ, start_response):
        environ['wikkid.user'] = self.user
        req = Request(environ)
        resp = req.get_response(self.app)
        return resp(environ, start_response)


@implementer(IUserFactory)
class UserFactory(object):
    """Generate a user from local bazaar config."""

    def __init__(self, branch):
        """Use the user config from the branch."""
        config = branch.get_config()
        self.user = create_bzr_user_from_author_string(config.username())
        logger = logging.getLogger('wikkid')
        logger.info(
            'Using bzr identity: "%s", "%s"',
            self.user.display_name, self.user.email)

    def create(self, request):
        """Create a User."""
        return self.user


@implementer(IUser)
class User(BaseUser):
    """A user from the local bazaar config."""

    def __init__(self, email, display_name, committer_id):
        self.email = email
        self.display_name = display_name
        self.committer_id = committer_id



# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""View classes to control the rendering of the content."""

from wikkid.interface.resource import IMissingResource
from wikkid.view.base import BaseView
from wikkid.view.edit import BaseEditView
from wikkid.view.textfile import SaveNewTextContent


class BaseMissingView(BaseView):
    """A base view for +view and +listing.

    This view just makes the results actual 404s.
    """

    def make_response(self, body):
        response = super(BaseMissingView, self).make_response(body)
        response.status = "404 Not Found"
        return response


class MissingPage(BaseMissingView):
    """A wiki page that does not exist."""

    for_interface = IMissingResource
    name = 'view'
    is_default = True
    template = 'missing'

    @property
    def content(self):
        '%s Not found' % self.path


class MissingDirectory(BaseMissingView):
    """A wiki directory that does not exist."""

    for_interface = IMissingResource
    name = 'listing'
    template = 'missing-dir'

    @property
    def content(self):
        '%s Not found' % self.path


class NewWikiPage(BaseEditView):
    """Show the edit page with no existing content."""

    for_interface = IMissingResource

    @property
    def rev_id(self):
        return None

    @property
    def content(self):
        return ''


class SaveNewTextFile(SaveNewTextContent):

    for_interface = IMissingResource



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The dispatcher for wikkid views.

When this module is loaded, it will automagically load all the other views in
this directory.  The views inherit from the BaseView which has a metaclass
which registers the view with the dispatcher.
"""

import os

from breezy.urlutils import dirname, joinpath

from zope.interface import providedBy

from wikkid.context import ExecutionContext

# The view registry needs to map an Interface and a name to a class.
_VIEW_REGISTRY = {}


def get_view(obj, view_name, request, ec=ExecutionContext()):
    """Get the most relevant view for the object for the specified name.

    Iterate through the provided interfaces of the object and look in the view
    registry for a view.
    """
    interfaces = providedBy(obj)
    for interface in interfaces:
        try:
            klass = _VIEW_REGISTRY[(interface, view_name)]
            instance = klass(obj, request, ec)
            instance.initialize()
            return instance
        except KeyError:
            pass
    # For example, if someone asked for 'raw' view on a directory or binary
    # object.
    return None


def register_view(view_class):
    """Register the view."""
    interface = getattr(view_class, 'for_interface', None)
    view_name = getattr(view_class, 'name', None)
    default_view = getattr(view_class, 'is_default', False)

    if view_name is None or interface is None:
        # Don't register.
        return
    key = (interface, view_name)
    assert key not in _VIEW_REGISTRY, "key already registered: %r" % (key,)
    _VIEW_REGISTRY[key] = view_class
    if default_view:
        _VIEW_REGISTRY[(interface, None)] = view_class


def unregister_view(view_class):
    """Unregister the view."""
    interface = getattr(view_class, 'for_interface', None)
    view_name = getattr(view_class, 'name', None)
    default_view = getattr(view_class, 'is_default', False)

    if view_name is None or interface is None:
        # Don't register.
        return
    key = (interface, view_name)
    assert _VIEW_REGISTRY[key] is view_class, \
        "key registered with different class: %r: %r != %r" % (
            key, _VIEW_REGISTRY[key], view_class)
    del _VIEW_REGISTRY[key]
    if default_view:
        del _VIEW_REGISTRY[(interface, None)]


# We know that the controller (whatever that is going to end up being) will
# load this module to get the 'get_view' function.  None of the other view
# modules should be explicitly loaded anywhere else (possible exceptions may
# occur, so this isn't a hard rule).
#
# So... when this module is loaded, we want to load the other modules in the
# wikkid.view package so that when the classes are parsed, they register
# themselves with the view registry.

def load_view_modules():
    curr_dir = os.path.abspath(dirname(__file__))
    view_dir = joinpath(curr_dir, 'view')
    py_files = [
        filename for filename in os.listdir(view_dir)
        if filename.endswith('.py') and not filename.startswith('__')]
    for filename in py_files:
        __import__('wikkid.view.%s' % filename[:-3])


load_view_modules()



#
# Copyright (C) 2012 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A git filestore using Dulwich.

"""

import datetime
import mimetypes

from dulwich.objects import Blob, Tree, ZERO_SHA
from dulwich.object_store import tree_lookup_path
from dulwich.repo import Repo
from dulwich.walk import Walker
import posixpath
import stat

from zope.interface import implementer

from wikkid.filestore import FileExists, UpdateConflicts
from wikkid.interface.filestore import FileType, IFile, IFileStore


@implementer(IFileStore)
class FileStore(object):
    """A filestore that just uses an internal map to store data."""

    _encoding = 'utf-8'

    @classmethod
    def from_path(cls, path):
        return cls(Repo(path))

    def __init__(self, repo, ref=b'HEAD'):
        """Repo is a dulwich repository."""
        self.repo = repo
        self.ref = ref

    @property
    def store(self):
        return self.repo.object_store

    def _get_root(self, revision=None):
        if revision is None:
            try:
                revision = self.repo.refs[self.ref]
            except KeyError:
                revision = ZERO_SHA
        try:
            return (revision, self.repo[revision].tree)
        except KeyError:
            return None, None

    def get_file(self, path):
        """Return an object representing the file."""
        commit_id, root_id = self._get_root()
        if root_id is None:
            return None
        try:
            (mode, sha) = tree_lookup_path(
                self.store.__getitem__,
                root_id, path.encode(self._encoding))
        except KeyError:
            return None
        return File(
            self.store, mode, sha, path, commit_id, encoding=self._encoding)

    def update_file(self, path, content, author, parent_revision,
                    commit_message=None):
        """The `author` is updating the file at `path` with `content`."""
        commit_id, root_id = self._get_root()
        if root_id is None:
            root_tree = Tree()
        else:
            root_tree = self.store[root_id]
        # Find all tree objects involved
        tree = root_tree
        trees = [root_tree]
        elements = path.strip(posixpath.sep).split(posixpath.sep)
        for el in elements[:-1]:
            try:
                (mode, sha) = tree[el.encode(self._encoding)]
            except KeyError:
                tree = Tree()
            else:
                if not stat.S_ISDIR(mode):
                    raise FileExists(
                        "File %s exists and is not a directory" % el)
                tree = self.store[sha]
            trees.append(tree)
        if elements[-1] in tree:
            (old_mode, old_sha) = tree[elements[-1]]
            if stat.S_ISDIR(old_mode):
                raise FileExists("File %s exists and is a directory" % path)
            if old_sha != parent_revision and parent_revision is not None:
                raise UpdateConflicts(
                    "File conflict %s != %s" % (
                        old_sha, parent_revision), old_sha)
        if not isinstance(content, bytes):
            raise TypeError(content)
        blob = Blob.from_string(content)
        child = (stat.S_IFREG | 0o644, blob.id)
        self.store.add_object(blob)
        assert len(trees) == len(elements)
        for tree, name in zip(reversed(trees), reversed(elements)):
            assert name != ""
            tree[name.encode(self._encoding)] = child
            self.store.add_object(tree)
            child = (stat.S_IFDIR, tree.id)
        if commit_message is None:
            commit_message = ""
        if author is not None:
            author = author.encode(self._encoding)
        self.repo.do_commit(
            ref=self.ref, message=commit_message.encode(self._encoding),
            author=author, tree=child[1])

    def list_directory(self, directory_path):
        """Return a list of File objects for in the directory path.

        If the path doesn't exist, returns None.  If the path exists but is
        empty, an empty list is returned.  Otherwise a list of File objects in
        that directory.
        """
        if directory_path is None:
            directory_path = ''
        else:
            directory_path = directory_path.strip(posixpath.sep)
        commit_id, root_id = self._get_root()
        if directory_path == '':
            sha = root_id
            mode = stat.S_IFDIR
        else:
            if root_id is None:
                return None
            try:
                (mode, sha) = tree_lookup_path(
                    self.store.__getitem__,
                    root_id, directory_path.encode(self._encoding))
            except KeyError:
                return None
        if mode is not None and stat.S_ISDIR(mode):
            ret = []
            for (name, mode, sha) in self.store[sha].items():
                ret.append(
                    File(self.store, mode, sha,
                         posixpath.join(
                             directory_path, name.decode(self._encoding)),
                         commit_id, encoding=self._encoding))
            return ret
        else:
            return None


@implementer(IFile)
class File(object):
    """A Git file object."""

    def __init__(self, store, mode, sha, path, commit_sha, encoding):
        self.store = store
        self.encoding = encoding
        self.mode = mode
        self.sha = sha
        self.path = path
        self.commit_sha = commit_sha
        self.base_name = posixpath.basename(path)
        self.mimetype = mimetypes.guess_type(self.base_name)[0]

    @property
    def file_type(self):
        """Work out the filetype based on the mimetype if possible."""
        if self._is_directory:
            return FileType.DIRECTORY
        else:
            if self.mimetype is None:
                binary = self._is_binary
            else:
                binary = not self.mimetype.startswith('text/')
            if binary:
                return FileType.BINARY_FILE
            else:
                return FileType.TEXT_FILE

    def get_content(self):
        o = self.store[self.sha]
        if isinstance(o, Blob):
            return o.data
        else:
            return None

    @property
    def _is_directory(self):
        return stat.S_ISDIR(self.mode)

    @property
    def _is_binary(self):
        return b'\0' in self.get_content()

    def _get_last_modified_commit(self):
        walker = Walker(
            self.store, include=[self.commit_sha],
            paths=[self.path.encode(self.encoding)])
        return next(iter(walker)).commit

    @property
    def last_modified_in_revision(self):
        return self.sha

    @property
    def last_modified_by(self):
        return self._get_last_modified_commit().author.decode(self.encoding)

    @property
    def last_modified_date(self):
        c = self._get_last_modified_commit()
        return datetime.datetime.utcfromtimestamp(c.commit_time)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the wikkid.filestore.bzr.FileStore."""

from textwrap import dedent

from breezy.tests import TestCaseWithTransport

from wikkid.filestore import UpdateConflicts
from wikkid.filestore.bzr import (
    BranchFileStore,
    FileStore,
    )
from wikkid.tests import ProvidesMixin
from wikkid.tests.filestore import TestFileStore


class TestBzrFileStore(TestCaseWithTransport, ProvidesMixin, TestFileStore):
    """Tests for the bzr filestore and files."""

    def make_filestore(self, contents=None):
        tree = self.make_branch_and_tree('.')
        if contents:
            self.build_tree_contents(contents)
        tree.smart_add(['.'])
        tree.commit(message='Initial commit', authors=['test@example.com'])
        return FileStore(tree)

    def test_conflicts(self):
        filestore = self.make_filestore(
            [('test.txt', b'one line of content\n')])
        f = filestore.get_file('test.txt')
        base_rev = f.last_modified_in_revision
        # First update succeeds.
        filestore.update_file(
            'test.txt',
            b'different line\n',
            'Test Author <test@example.com>',
            base_rev)
        # Second update with same base revision should raise an exception.
        conflicts = self.assertRaises(
            UpdateConflicts,
            filestore.update_file,
            'test.txt',
            b'also change the first line\n',
            'Test Author <test@example.com>',
            base_rev)
        curr = filestore.get_file('test.txt')
        self.assertEqual(
            curr.last_modified_in_revision,
            conflicts.basis_rev)
        self.assertEqual(dedent("""\
            <<<<<<<
            also change the first line
            =======
            different line
            >>>>>>>
            """).encode(), conflicts.content)

    def test_conflicts_dos_line_endings(self):
        filestore = self.make_filestore(
            [('test.txt', b'one line of content\r\n')])
        f = filestore.get_file('test.txt')
        base_rev = f.last_modified_in_revision
        # First update succeeds.
        filestore.update_file(
            'test.txt',
            b'different line\r\n',
            'Test Author <test@example.com>',
            base_rev)
        # Second update with same base revision should raise an exception.
        conflicts = self.assertRaises(
            UpdateConflicts,
            filestore.update_file,
            'test.txt',
            b'also change the first line\r\n',
            'Test Author <test@example.com>',
            base_rev)
        curr = filestore.get_file('test.txt')
        self.assertEqual(
            curr.last_modified_in_revision,
            conflicts.basis_rev)
        self.assertEqual(
            b"<<<<<<<\r\n"
            b"also change the first line\r\n"
            b"=======\r\n"
            b"different line\r\n"
            b">>>>>>>\r\n",
            conflicts.content)

    def test_line_endings_unix(self):
        # If the underlying file has unix file endings, and the post param has
        # dos endings, the post param is converted before the merge is
        # attempted.
        filestore = self.make_filestore(
            [('test.txt', b'several\nlines\nof\ncontent')])
        f = filestore.get_file('test.txt')
        base_rev = f.last_modified_in_revision
        # Updating the text with dos endings doesn't convert the file.
        filestore.update_file(
            'test.txt',
            b'several\r\nslightly different lines\r\nof\r\ncontent',
            'Test Author <test@example.com>',
            base_rev)
        curr = filestore.get_file('test.txt')
        # New line added too.
        self.assertEqual(
            b'several\nslightly different lines\nof\ncontent\n',
            curr.get_content())

    def test_line_endings_new_file(self):
        # Line endings for new files default to '\n'.
        filestore = self.make_filestore()
        filestore.update_file(
            'new-file.txt',
            b'some\r\ndos\r\nline\r\nendings',
            'Test Author <test@example.com>',
            None)
        curr = filestore.get_file('new-file.txt')
        # A new line is added to the end too.
        self.assertEqual(
            b'some\ndos\nline\nendings\n',
            curr.get_content())

    def test_empty(self):
        # Empty files do not have line endings, but they can be saved
        # nonetheless.
        filestore = self.make_filestore(
            [('test.txt', b'several\nlines\nof\ncontent')])
        f = filestore.get_file('test.txt')
        base_rev = f.last_modified_in_revision
        filestore.update_file(
            'test.txt', b'', 'Test Author <test@example.com>', base_rev)
        curr = filestore.get_file('test.txt')
        self.assertEqual(b'', curr.get_content())


class TestBranchFileStore(TestBzrFileStore):

    def make_filestore(self, contents=None):
        tree = self.make_branch_and_tree('.')
        if contents:
            self.build_tree_contents(contents)
        tree.smart_add(['.'])
        tree.commit(message='Initial commit', authors=['test@example.com'])
        return BranchFileStore(tree.branch)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests the breadcrumbs for the views."""

from wikkid.tests.factory import ViewTestCase
from wikkid.tests.fakes import TestRequest, TestUser


class TestBreadcrumbs(ViewTestCase):
    """The breadcrumbs lead the user back home."""

    # Don't test the breadcrumbs for the root object directly here as it will
    # redirect to Home.  TODO: add a test browser test for this.

    def setUp(self):
        super(TestBreadcrumbs, self).setUp()
        self.user = TestUser('test@example.com', 'Test User')
        self.request = TestRequest()

    def assertBreadcrumbs(self, view, expected):
        """Make sure the breadcrumbs from view are the expected ones."""
        crumbs = [(crumb.title, crumb.path) for crumb in view.breadcrumbs]
        self.assertEqual(expected, crumbs)

    def test_home_missing(self):
        # If the Home page is selected, but there is no content, the
        # breadcrumb is still Home.
        factory = self.make_factory()
        view = self.get_view(factory, '/Home')
        self.assertBreadcrumbs(
            view,
            [('Home', '/Home')])

    def test_page_missing(self):
        # If the page is at the root of the tree, but isn't home, then the
        # first breadcrumb is home, and the second is the page.
        factory = self.make_factory()
        view = self.get_view(factory, '/SamplePage')
        self.assertBreadcrumbs(
            view,
            [('Home', '/Home'),
             ('Sample Page', '/SamplePage')])

    def test_nested_page_missing(self):
        # If the Home page is selected, but there is no content, the
        # breadcrumb is still Home.
        factory = self.make_factory()
        view = self.get_view(factory, '/SamplePage/SubPage/Next')
        self.assertBreadcrumbs(
            view,
            [('Home', '/Home'),
             ('Sample Page', '/SamplePage'),
             ('Sub Page', '/SamplePage/SubPage'),
             ('Next', '/SamplePage/SubPage/Next')])

    def test_source_file_missing(self):
        # If a non-wiki style name is selected, the breadcrumbs are as a wiki
        # page.
        factory = self.make_factory()
        view = self.get_view(factory, '/wikkid/views/base.py')
        self.assertBreadcrumbs(
            view,
            [('Home', '/Home'),
             ('wikkid', '/wikkid'),
             ('views', '/wikkid/views'),
             ('base.py', '/wikkid/views/base.py')])

    def test_source_file_existing(self):
        # If the Home page is selected, but there is no content, the
        # breadcrumb is still Home.
        factory = self.make_factory([
                ('wikkid/views/base.py', 'A python file'),
                ])
        view = self.get_view(factory, '/wikkid/views/base.py')
        self.assertBreadcrumbs(
            view,
            [('Home', '/Home'),
             ('wiki root', '/+listing'),
             ('wikkid', '/wikkid/+listing'),
             ('views', '/wikkid/views/+listing'),
             ('base.py', '/wikkid/views/base.py')])

    def test_directory_breadcrumbs_root(self):
        # Directory breadcrumbs start with Home, and then list the
        # directories, where the urls for the directories are the listing
        # urls.
        factory = self.make_factory()
        view = self.get_view(factory, '/', 'listing')
        self.assertBreadcrumbs(
            view,
            [('Home', '/Home'),
             ('wiki root', '/+listing')])

    def test_directory_breadcrumbs_nested(self):
        # For each directory after the root, a listing crumb is added.
        # Names are not wiki expanded.
        factory = self.make_factory([
                ('SomePage/SubPage/Nested.txt', 'some text')])
        view = self.get_view(factory, '/SomePage/SubPage', 'listing')
        self.assertBreadcrumbs(
            view,
            [('Home', '/Home'),
             ('wiki root', '/+listing'),
             ('SomePage', '/SomePage/+listing'),
             ('SubPage', '/SomePage/SubPage')])

    def test_directory_breadcrumbs_nested_with_script_name(self):
        # For each directory after the root, a listing crumb is added.
        # Names are not wiki expanded.
        factory = self.make_factory([
                ('SomePage/SubPage/Nested.txt', 'some text')])
        view = self.get_view(
            factory, '/SomePage/SubPage', 'listing', '/p/wiki')
        self.assertBreadcrumbs(
            view,
            [('Home', '/p/wiki/Home'),
             ('wiki root', '/p/wiki/+listing'),
             ('SomePage', '/p/wiki/SomePage/+listing'),
             ('SubPage', '/p/wiki/SomePage/SubPage')])



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the wikkid.model.factory module."""

from wikkid.interface.resource import (
    IBinaryFile,
    IDirectoryResource,
    IMissingResource,
    IRootResource,
    ISourceTextFile,
    IWikiTextFile,
    )
from wikkid.tests.factory import FactoryTestCase

# TODO: make a testing filestore that can produce either a volatile filestore
# or a bzr filestore.


class TestFactoryGetResourceAtPath(FactoryTestCase):
    """Test the get_resource_at_path method of the Factory class."""

    def test_get_resource_at_path_root_no_content(self):
        # If the root file is selected, and there is no content, there is no
        # read_filename or file_resource, but there is a write_filename for
        # the default home page.
        factory = self.make_factory()
        info = factory.get_resource_at_path('/')
        self.assertProvides(info, IRootResource)
        self.assertEqual('/', info.path)
        self.assertIs(None, info.get_dir_name())
        self.assertIs(None, info.dir_resource)
        self.assertFalse(info.has_home_page)

    def test_get_resource_at_path_root_has_page(self):
        # If the root file is selected, and there is a home page, this is
        # returned.
        factory = self.make_factory([
                ('Home.txt', b'the home page'),
                ])
        info = factory.get_resource_at_path('/')
        self.assertProvides(info, IRootResource)
        self.assertEqual('/', info.path)
        self.assertIs(None, info.get_dir_name())
        self.assertIs(None, info.dir_resource)
        self.assertEqual('Home.txt', info.file_resource.path)
        self.assertTrue(info.has_home_page)

    def test_get_resource_at_path_missing_file(self):
        # A missing file as no read filename, nor resource, but does have
        # a write file name.
        factory = self.make_factory()
        info = factory.get_resource_at_path('/missing-file')
        self.assertProvides(info, IMissingResource)
        self.assertEqual('/missing-file', info.path)
        self.assertEqual('missing-file.txt', info.write_filename)
        self.assertIs(None, info.file_resource)
        self.assertIs(None, info.dir_resource)

    def test_get_resource_at_path_existing_file_no_suffix(self):
        # If a file is requested that exists but has no suffix, it is returned
        # unaltered, and is interpreted as a wiki page.
        factory = self.make_factory([
                ('README', b'A readme file'),
                ])
        info = factory.get_resource_at_path('/README')
        self.assertProvides(info, IWikiTextFile)
        self.assertEqual('/README', info.path)
        self.assertEqual('README', info.write_filename)
        self.assertEqual('README', info.file_resource.path)
        self.assertIs(None, info.dir_resource)

    def test_get_resource_at_path_missing_file_not_text(self):
        # A missing file is requested, but has a suffix, don't attempt to add
        # a .txt to it.
        factory = self.make_factory()
        info = factory.get_resource_at_path('/missing-file.cpp')
        self.assertProvides(info, IMissingResource)
        self.assertEqual('/missing-file.cpp', info.path)
        self.assertEqual('missing-file.cpp', info.write_filename)
        self.assertIs(None, info.file_resource)
        self.assertIs(None, info.dir_resource)

    def test_get_resource_at_path_not_text(self):
        # If a file is reqeusted that has a suffix that isn't text,
        # a source text file object is returned,
        factory = self.make_factory([
                ('sample.cpp', b'A c++ file'),
                ])
        info = factory.get_resource_at_path('/sample.cpp')
        self.assertProvides(info, ISourceTextFile)
        self.assertEqual('/sample.cpp', info.path)
        self.assertEqual('sample.cpp', info.write_filename)
        self.assertEqual('sample.cpp', info.file_resource.path)
        self.assertIs(None, info.dir_resource)

    def test_get_resource_at_path_binary_file(self):
        # If a file exists and is binary, a binary file is returned.
        factory = self.make_factory([
                ('sample.png', b'A picture'),
                ])
        info = factory.get_resource_at_path('/sample.png')
        self.assertProvides(info, IBinaryFile)
        self.assertEqual('/sample.png', info.path)
        self.assertEqual('sample.png', info.write_filename)
        self.assertEqual('sample.png', info.file_resource.path)
        self.assertIs(None, info.dir_resource)

    def test_get_resource_at_path_directory(self):
        # A directory without a matching text file doesn't have a
        # read_filename either, but does have a dir_resource.
        factory = self.make_factory([
                ('SomeDir/', None),
                ])
        info = factory.get_resource_at_path('/SomeDir')
        self.assertProvides(info, IDirectoryResource)
        self.assertEqual('/SomeDir', info.path)
        self.assertEqual('SomeDir.txt', info.write_filename)
        self.assertIs(None, info.file_resource)
        self.assertEqual('SomeDir', info.get_dir_name())

    def test_get_resource_at_path_directory_with_page(self):
        # A directory with a matching text file has a text resource and a dir
        # resource.
        factory = self.make_factory([
                ('SomeDir/', None),
                ('SomeDir.txt', b'Some content'),
                ])
        info = factory.get_resource_at_path('/SomeDir')
        self.assertProvides(info, IWikiTextFile)
        self.assertEqual('/SomeDir', info.path)
        self.assertEqual('SomeDir.txt', info.write_filename)
        self.assertEqual('SomeDir.txt', info.file_resource.path)
        self.assertIsNot('SomeDir', info.dir_resource)

    def test_get_resource_at_path_page_no_directory(self):
        # A directory with a matching text file has a text resource and a dir
        # resource.
        factory = self.make_factory([
                ('SomeDir.txt', b'Some content'),
                ])
        info = factory.get_resource_at_path('/SomeDir')
        self.assertProvides(info, IWikiTextFile)
        self.assertEqual('/SomeDir', info.path)
        self.assertEqual('SomeDir.txt', info.write_filename)
        self.assertEqual('SomeDir.txt', info.file_resource.path)
        self.assertIs(None, info.dir_resource)

    def test_get_resource_at_path_subdirs_missing_file(self):
        # The path info reflects the request path.
        factory = self.make_factory()
        info = factory.get_resource_at_path('/a/b/c/d')
        self.assertProvides(info, IMissingResource)
        self.assertEqual('/a/b/c/d', info.path)
        self.assertEqual('a/b/c/d.txt', info.write_filename)
        self.assertIs(None, info.file_resource)
        self.assertIs(None, info.dir_resource)

    def test_get_resource_at_path_subdirs_existing_file(self):
        # The path info reflects the request path.
        factory = self.make_factory([
                ('a/b/c/d.txt', b'a text file'),
                ])
        info = factory.get_resource_at_path('/a/b/c/d')
        self.assertProvides(info, IWikiTextFile)
        self.assertEqual('/a/b/c/d', info.path)
        self.assertEqual('a/b/c/d.txt', info.write_filename)
        self.assertEqual('a/b/c/d.txt', info.file_resource.path)
        self.assertIs(None, info.dir_resource)

    def test_get_resource_at_path_subdirs_existing_file_and_dir(self):
        # The path info reflects the request path.
        factory = self.make_factory([
                ('a/b/c/d.txt', b'a text file'),
                ('a/b/c/d/e.txt', b'another text file'),
                ])
        info = factory.get_resource_at_path('/a/b/c/d')
        self.assertProvides(info, IWikiTextFile)
        self.assertEqual('/a/b/c/d', info.path)
        self.assertEqual('a/b/c/d.txt', info.write_filename)
        self.assertEqual('a/b/c/d.txt', info.file_resource.path)
        self.assertEqual('a/b/c/d', info.dir_resource.path)


class TestFactoryGetPreferredPath(FactoryTestCase):
    """Tests for the get_preferred_path method of the Factory class."""

    def test_home_preferred(self):
        factory = self.make_factory()
        self.assertEqual('/Home', factory.get_preferred_path('/'))
        self.assertEqual('/Home', factory.get_preferred_path('/Home'))
        self.assertEqual('/Home', factory.get_preferred_path('/Home.txt'))

    def test_default_preferred(self):
        factory = self.make_factory()
        factory.DEFAULT_PATH = 'FrontPage'
        self.assertEqual(
            '/FrontPage', factory.get_preferred_path('/'))
        self.assertEqual(
            '/FrontPage', factory.get_preferred_path('/FrontPage'))
        self.assertEqual(
            '/FrontPage', factory.get_preferred_path('/FrontPage.txt'))

    def test_image_preferred(self):
        factory = self.make_factory()
        self.assertEqual(
            '/foo/bar.jpg', factory.get_preferred_path('/foo/bar.jpg'))

    def test_text_preferred(self):
        factory = self.make_factory()
        self.assertEqual('/README', factory.get_preferred_path('/README'))
        self.assertEqual('/a.b.txt', factory.get_preferred_path('/a.b.txt'))
        self.assertEqual('/a', factory.get_preferred_path('/a.txt'))
        self.assertEqual('/a/b', factory.get_preferred_path('/a/b.txt'))



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""For packages that are not adequately packaged."""



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The binary resource class.

A binary resource is a file that isn't text.  This is primarily guessed using
the mimetype library.
"""

from zope.interface import implementer

from wikkid.model.file import FileResource
from wikkid.interface.resource import IBinaryFile


@implementer(IBinaryFile)
class BinaryResource(FileResource):
    """A binary resource is a non-text file."""

    def __repr__(self):
        return "<BinaryResource '%s'>" % self.path



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Interfaces relating to filestores."""

from zope.interface import Attribute, Interface


class IFileStore(Interface):
    """A file store is how pages are accessed and updated.

    This interface defines methods for getting file contents, checking to see
    if the file exists, updating files, checking differences, and many other
    wiki type thiings.
    """

    def get_file(path):
        """Return an object representing the file at specified path."""

    def update_file(path, content, author, parent_revision,
                    commit_message=None):
        """Update a text file at the given path with the content.

        :param path: The path of the file.
        :param content: The content of the file.
        :param author: Who is doing the updating.
        :type author: String
        :param parent_revision: The revision that the user was editing when
           they made the changes.  For a new revision this parameter will be
           None.
        :type parent_revision: String
        :param commit_message: An optional commit message.  If one isn't
           provided, then some sensible default will be used.
        """

    def list_directory(path):
        """Return a list of IFile objects.

        Each of the IFile objects will be directly in the directory specified
        by the path.

        If the specified path is None, the files in the root directory of the
        branch are returned.

        If the specified path doesn't exist, None is returned.

        If the specified path exists but has no files, an empty list is
        returned.
        """


class FileType(object):
    """Package lazr.enum and use an Enumerated Type."""

    MISSING = 1  # The file at the address does not exist.
    WIKI_PAGE = 2  # The resource is a wiki page.
    DIRECTORY = 3  # The resource is a directory.
    TEXT_FILE = 4  # A text file that isn't a wiki page.
    BINARY_FILE = 5  # A (most likely) binary file.


class IFile(Interface):
    """A file from the file store."""

    path = Attribute(
        "The full path of the page with respect to the root of the "
        "file store.")

    base_name = Attribute("The last part of the path.")

    file_type = Attribute("Soon to be a Choice with a lazr.enum.")

    mimetype = Attribute(
        "The guessed mimetype for the file. Directories don't have a "
        "mimetype.")

    last_modified_in_revision = Attribute(
        "The revision id of the last revision that this file was "
        "modified in.")

    last_modified_by = Attribute("The person who last modified the file.")

    last_modified_date = Attribute(
        'The timestamp of the revision that last modified the file. '
        'This is a naive datetime object in UTC.')

    def get_content():
        """Get the contents of the file.

        :return: None if the file doesn't yet exist, or u'' if the file is
            empty, otherwise the unicode content of the file.
        """



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The skins that are provided for Wikkid Wiki.

The are the following defined skins:
  default
"""



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Utility methods for wikkid views."""

import re


WIKI_PAGE = re.compile('^([A-Z]+[a-z]*)+$')
WIKI_PAGE_ELEMENTS = re.compile('([A-Z][a-z]+)')


def expand_wiki_name(name):
    """A wiki name like 'FrontPage' is expanded to 'Front Page'.

    Names that don't match wiki names are unaltered.
    """
    if WIKI_PAGE.match(name):
        name_parts = [
            part for part in WIKI_PAGE_ELEMENTS.split(name) if part]
        return ' '.join(name_parts)
    else:
        return name


def title_for_filename(filename):
    """Generate a title based on the basename of the file object."""
    if filename.endswith('.txt'):
        return expand_wiki_name(filename[:-4])
    else:
        return expand_wiki_name(filename)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The source text class.

A source text file is a text file that isn't a wiki file.
"""

from zope.interface import implementer

from wikkid.model.file import FileResource
from wikkid.interface.resource import ITextFile


@implementer(ITextFile)
class TextFile(FileResource):
    """A text file that isn't a wiki page."""

    def __repr__(self):
        return "<TextFile '%s'>" % self.path



#! /usr/bin/python
#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Add the parent directory to the python path to run the scripts."""

import sys
import os.path

current_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(current_dir)
sys.path.insert(0, parent_dir)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The root resource class.

The root resource represents the object at the root of the wiki path.

Currently this just refers to '/', but it is expected that at some stage in
the not too distant future the server will support a wiki root where it is not
the root path.
"""

from zope.interface import implementer

from wikkid.model.directory import DirectoryResource
from wikkid.interface.resource import IRootResource


@implementer(IRootResource)
class RootResource(DirectoryResource):
    """The root of the wiki.

    Some special wiki views hang off the root resource and not others.  A root
    resource is also a directory resource where the directory is the root of
    the filesystem.
    """

    def get_dir_name(self):
        return None

    @property
    def has_home_page(self):
        return self.file_resource is not None

    def __repr__(self):
        return "<RootResource '/'>"

    @property
    def preferred_path(self):
        return '/'



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the wikkid.formatter.markdownformatter module."""

from textwrap import dedent

from bs4 import BeautifulSoup

# The markdown formatter is optional.  If the MarkdownFormatter is not
# available, the tests are skipped.
try:
    from wikkid.formatter.markdownformatter import MarkdownFormatter
    has_markdown = True
except ImportError:
    has_markdown = False
from wikkid.tests import TestCase


class TestMarkdownFormatter(TestCase):
    """Tests for the markdown formatter."""

    def setUp(self):
        TestCase.setUp(self)
        if has_markdown:
            self.formatter = MarkdownFormatter()
        else:
            self.skip('markdown formatter not available')

    def test_simple_headings(self):
        # A simple heading and a paragraph.
        text = dedent("""\
            Heading 1
            =========

            Heading 2
            ---------

            Simple sentence.
            """)
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result, features="lxml")
        self.assertEqual('Heading 1', soup.h1.string)
        self.assertEqual('Heading 2', soup.h2.string)
        # We don't care about whitespace on <p> tags, and markdown inserts
        # newlines.
        self.assertEqual('Simple sentence.', soup.p.string.strip())

    def test_detailed_headings(self):
        text = dedent("""\
            # Heading 1

            ## Heading 2

            ### Heading 3

            #### Heading 4

            ##### Heading 5

            ###### Heading 6""")
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result, features="lxml")
        self.assertEqual('Heading 1', soup.h1.string)
        self.assertEqual('Heading 2', soup.h2.string)
        self.assertEqual('Heading 3', soup.h3.string)
        self.assertEqual('Heading 4', soup.h4.string)
        self.assertEqual('Heading 5', soup.h5.string)
        self.assertEqual('Heading 6', soup.h6.string)

    def test_inline_link(self):
        # A paragraph containing a wiki word.
        text = "A link to the [FrontPage](http://127.0.0.1) helps."
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result, features="lxml")
        self.assertEqual('http://127.0.0.1', soup.a['href'])

    def test_reference_link(self):
        # A paragraph containing a wiki word.
        text = dedent("""\
            This is [an example][id] reference-style link.

            [id]: http://127.0.0.1
            """)
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result, features="lxml")
        self.assertEqual('http://127.0.0.1', soup.a['href'])

    def test_emphasis(self):
        texts = ('We can have *emphasis* and **strong** as well!',
                 'We can have _emphasis_ and __strong__ as well!')
        for text in texts:
            result = self.formatter.format('filename', text)
            soup = BeautifulSoup(result, features="lxml")
            self.assertEqual('emphasis', soup.em.string)
            self.assertEqual('strong', soup.strong.string)

    def test_blockquote(self):
        text = dedent("""\
            > This is a block quoted paragraph
            > that spans multiple lines.
            """)
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result, features="lxml")
        self.assertTrue(soup.blockquote is not None)

    def test_lists(self):
        text = dedent("""\
            Some Text.

             * UL 1
             * UL 2
             * UL 3

            Some More Text!

             1. OL 1
             2. OL 2
             7. OL 3
             """)
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result, features="lxml")

        self.assertIsNot(None, soup.ul)
        ulNodes = soup.ul.findAll('li')
        self.assertEqual(
            ['UL 1', 'UL 2', 'UL 3'],
            [node.string.strip() for node in ulNodes])

        self.assertIsNot(None, soup.ol)
        olNodes = soup.ol.findAll('li')
        self.assertEqual(
            ['OL 1', 'OL 2', 'OL 3'],
            [node.string.strip() for node in olNodes])

    def test_code_blocks(self):
        text = dedent("""\
            Some Normal Text.

                Some Code inside pre tags

            More Normal text.
            """)
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result, features="lxml")
        self.assertEqual(
            soup.pre.code.string.strip(), 'Some Code inside pre tags')

    def test_hr(self):
        # test different HR types:
        texts = ('* * *',
                 '***',
                 '*********',
                 '- - - -',
                 '------------------')
        for text in texts:
            result = self.formatter.format('filename', text)
            soup = BeautifulSoup(result, features="lxml")
            self.assertIsNot(None, soup.hr)

    def test_code(self):
        text = 'use the `printf()` function'
        result = self.formatter.format('filename', text)
        soup = BeautifulSoup(result, features="lxml")
        self.assertEqual('printf()', soup.code.string)

    def test_unicode(self):
        # Test the unicode support of the markdown formatter.
        text = u'\N{SNOWMAN}'
        result = self.formatter.format('format', text)
        soup = BeautifulSoup(result, features="lxml")
        self.assertEqual(soup.p.string.strip(), text)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The wikkid tests and test only code."""

import unittest

import testtools
from zope.interface.verify import verifyObject


class ProvidesMixin(object):

    def assertProvides(self, obj, interface):
        """Assert 'obj' correctly provides 'interface'."""
        self.assertTrue(
            interface.providedBy(obj),
            "%r does not provide %r." % (obj, interface))
        self.assertTrue(
            verifyObject(interface, obj),
            "%r claims to provide %r but does not do so correctly."
            % (obj, interface))


class TestCase(testtools.TestCase, ProvidesMixin):
    """Add some zope interface helpers."""


def test_suite():
    packages = [
        'formatters',
        'views',
        ]
    names = [
        'app',
        'context',
        'model_factory',
        'volatile_filestore',
        'bzr_filestore',
        'bzr_user',
        'git_filestore',
        'view_dispatcher',
        'model',
        ]
    module_names = ['wikkid.tests.test_' + name for name in names]
    loader = unittest.TestLoader()
    suite = loader.loadTestsFromNames(module_names)
    for pkgname in packages:
        pkg = __import__(
            'wikkid.tests.' + pkgname, globals(), locals(), ['test_suite'])
        suite.addTests(pkg.test_suite())
    return suite



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the wikkid.filestore.volatile.FileStore."""

from wikkid.tests import TestCase
from wikkid.tests.filestore import TestFileStore
from wikkid.filestore.volatile import FileStore


class TestVolatileFileStore(TestCase, TestFileStore):
    """Tests for the volatile filestore and files."""

    def make_filestore(self, contents=None):
        return FileStore(contents)



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""View classes to control the rendering of text pages."""

import logging

from webob.exc import HTTPSeeOther

from wikkid.filestore import UpdateConflicts
from wikkid.interface.resource import ITextFile
from wikkid.view.edit import BaseEditView
from wikkid.view.wiki import format_content


class EditTextFile(BaseEditView):
    """The page shows the text content in a large edit field."""

    for_interface = ITextFile

    @property
    def rev_id(self):
        return self.context.last_modified_in_revision

    @property
    def content(self):
        # We want to pass unicode to the view.
        byte_string = self.context.get_bytes()
        try:
            return byte_string.decode('utf-8')
        except UnicodeDecodeError:
            try:
                return byte_string.decode('latin-1')
            except UnicodeDecodeError:
                return byte_string.decode('ascii', 'replace')


class SaveNewTextContent(BaseEditView):
    """Update the text of a file."""

    name = 'save'

    def _render(self, skin):
        """Save the text file.

        If it conflicts, render the edit, otherwise render the page (ideally
        redirect back to the plain page.
        """
        # TODO: barf on a GET
        # TODO: barf if there is no user.
        params = self.request.params
        content = params['content']
        description = params['description']
        rev_id = (
            params['rev-id'].encode('utf-8') if 'rev-id' in params else None)
        preview = params.get('preview', None)
        if preview is not None:
            self.rev_id = rev_id
            self.description = description
            self.content = content
            default_format = self.execution_context.default_format
            self.preview_content = format_content(
                content, self.context.base_name, default_format)
        else:
            try:
                self.context.put_bytes(
                    content.encode('utf-8'), self.user.committer_id, rev_id,
                    description)

                location = self.canonical_url(self.context)
                raise HTTPSeeOther(location=location)
            except UpdateConflicts as e:
                # Show the edit page again.
                logger = logging.getLogger('wikkid')
                logger.info('Conflicts detected: \n%r\n', e.content)
                self.rev_id = e.basis_rev
                self.content = e.content
                self.message = "Conflicts detected during merge."

        self.description = description
        return super(SaveNewTextContent, self)._render(skin)


class UpdateTextFile(SaveNewTextContent):

    for_interface = ITextFile



# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The formatter registry is able to return formatters by name."""

import re

from wikkid.formatter.pygmentsformatter import PygmentsFormatter
from wikkid.formatter.restformatter import RestructuredTextFormatter

# Both textile and markdown are optional.
try:
    from wikkid.formatter.markdownformatter import MarkdownFormatter
    has_markdown = True
except ImportError:
    has_markdown = False
try:
    from wikkid.formatter.textileformatter import TextileFormatter
    has_textile = True
except ImportError:
    has_textile = False


class FormatterRegistry(object):
    """Has a dictionary of formatters based on name."""

    def __init__(self):
        self.formatters = {
            'pygments': PygmentsFormatter(),
            'rest': RestructuredTextFormatter(),
            }
        if has_markdown:
            self.formatters['markdown'] = MarkdownFormatter()
        if has_textile:
            self.formatters['textile'] = TextileFormatter()

    def __getitem__(self, formatter):
        return self.formatters[formatter]


formatter_registry = FormatterRegistry()


def get_formatter(name):
    """Get a formatter by name."""
    return formatter_registry[name]


FORMAT_MATCHER = re.compile(r'^#\s+(\w+).*$')


def get_wiki_formatter(content: str, default_formatter: str) -> str:
    """Choose a wiki formatter based on the first line of content.

    Args:
      content: The content of the file.
      default_formatter: The name of the default formatter.

    The first line of the content may specify a formatter using the form:
    # formatter-name

    For example:
    # pygments

    The first line must start with a # and the first word must specify
    a formatter name.  If niether of those match, the default_formatter
    is returned.  If the default_formatter doesn't exist, a key error
    is raised.
    """
    end_of_line = content.find('\n')
    match = FORMAT_MATCHER.match(content[:end_of_line])
    if match is not None:
        try:
            name = match.group(1).lower()
            formatter = formatter_registry[name]
            return content[end_of_line + 1:], formatter
        except KeyError:
            # Fall through to returning the default.
            pass
    return content, formatter_registry[default_formatter]



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A bzr backed filestore."""

from datetime import datetime
import logging

from zope.interface import implementer
from merge3 import Merge3

from breezy.bzr.generate_ids import gen_file_id
from breezy.errors import BinaryFile
from breezy.osutils import splitpath, split_lines
from breezy.revision import NULL_REVISION
from breezy.textfile import check_text_lines
from breezy.transform import FinalPaths, MalformedTransform
from breezy.urlutils import basename, dirname, joinpath

from wikkid.filestore import FileExists, UpdateConflicts
from wikkid.filestore.basefile import BaseFile
from wikkid.interface.filestore import FileType, IFile, IFileStore


def normalize_line_endings(content, ending=b'\n'):
    return ending.join(content.splitlines())


def get_line_ending(lines):
    """Work out the line ending used in lines."""
    if len(lines) == 0:
        return b'\n'
    first = lines[0]
    if first.endswith(b'\r\n'):
        return b'\r\n'
    # Default to \n if there are no line endings.
    return b'\n'


def get_commit_message(commit_message):
    if commit_message is None or commit_message.strip() == '':
        return 'No description of change given.'
    return commit_message


def normalize_content(content):
    # Default to simple '\n' line endings.
    content = normalize_line_endings(content)
    # Make sure the content ends with a new-line.  This makes
    # end of file conflicts nicer.
    if not content.endswith(b'\n'):
        content += b'\n'
    return content


def iter_paths(path):
    path_segments = splitpath(path)
    while len(path_segments) > 0:
        tail = path_segments.pop()
        if len(path_segments) == 0:
            yield '', tail
        else:
            yield joinpath(*path_segments), tail


def create_parents(tt, path, trans_id):
    prev_trans_id = trans_id
    for parent_path, tail in iter_paths(path):
        trans_id = tt.trans_id_tree_path(parent_path)
        if tt.tree_kind(trans_id) is not None:
            break
        tt.adjust_path(tail, trans_id, prev_trans_id)
        tt.create_directory(trans_id)
        tt.version_file(trans_id=trans_id, file_id=gen_file_id(basename(path)))
        prev_trans_id = trans_id


@implementer(IFileStore)
class FileStore(object):
    """Wraps a Bazaar branch to be a filestore."""

    def __init__(self, tree):
        self.tree = tree
        self.branch = tree.branch
        self.logger = logging.getLogger('wikkid')

    def basis_tree(self):
        return self.tree.basis_tree()

    def get_file(self, path):
        """Return an object representing the file at specified path."""
        if not self.tree.is_versioned(path):
            return None
        else:
            return File(self, path)

    def update_file(self, path, content, author, parent_revision,
                    commit_message=None):
        """Update the file at the specified path with the content.

        This is going to be really interesting when we need to deal with
        conflicts.
        """
        commit_message = get_commit_message(commit_message)
        if parent_revision is None:
            parent_revision = NULL_REVISION
        # Firstly we want to lock the tree for writing.
        with self.tree.lock_write():
            # Look to see if the path is there.  If it is then we are doing an
            # update.  If it isn't we are doing an add.
            if self.tree.is_versioned(path):
                # What if a parent_revision hasn't been set?
                self._update_file(
                    path, content, author, parent_revision,
                    commit_message)
            else:
                self._add_file(path, content, author, commit_message)

    def _ensure_directory_or_nonexistant(self, dir_path):
        """Ensure the dir_path defines a directory or doesn't exist.

        Walk up the dir_path and make sure that the path either doesn't exist
        at all, or is a directory.  The purpose of this is to make sure we
        don't try to add a file in a directory where the directory has the
        same name as an existing file.
        """
        check = []
        while dir_path:
            check.append(dir_path)
            dir_path = dirname(dir_path)
        while len(check):
            f = self.get_file(check.pop())
            if f is not None:
                if not f.is_directory:
                    raise FileExists(
                        '%s exists and is not a directory' % f.path)

    def _add_file(self, path, content, author, commit_message):
        """Add a new file at the specified path with the content.

        Then commit this new file with the specified commit_message.
        """
        content = normalize_content(content)
        t = self.tree.controldir.root_transport
        # Get a transport for the path we want.
        self._ensure_directory_or_nonexistant(dirname(path))
        t = t.clone(dirname(path))
        t.create_prefix()
        # Put the file there.
        # TODO: UTF-8 encode text files?
        t.put_bytes(basename(path), content)
        self.tree.smart_add([t.local_abspath('.')])
        self.tree.commit(
            message=commit_message,
            authors=[author])

    def _get_final_text(self, content, f, parent_revision):
        current_rev = f.last_modified_in_revision
        wt = self.tree
        current_lines = wt.get_file_lines(f.path)
        basis = self.branch.repository.revision_tree(parent_revision)
        basis_lines = basis.get_file_lines(f.path)
        # need to break content into lines.
        ending = get_line_ending(current_lines)
        # If the content doesn't end with a new line, add one.
        new_lines = split_lines(content)
        # Look at the end of the first string.
        new_ending = get_line_ending(new_lines)
        if ending != new_ending:
            # I know this is horribly inefficient, but lets get it working
            # first.
            content = normalize_line_endings(content, ending)
            new_lines = split_lines(content)
        if len(new_lines) > 0 and not new_lines[-1].endswith(ending):
            new_lines[-1] += ending
        merge = Merge3(basis_lines, new_lines, current_lines)
        result = b''.join(merge.merge_lines())  # or merge_regions or whatever
        conflicted = (b'>>>>>>>' + ending) in result
        if conflicted:
            raise UpdateConflicts(result, current_rev)
        return result

    def _update_file(self, path, content, author, parent_revision,
                     commit_message):
        """Update an existing file with the content.

        This method merges the changes in based on the parent revision.
        """
        f = File(self, path)
        wt = self.tree
        with wt.lock_write():
            result = self._get_final_text(content, f, parent_revision)
            wt.controldir.root_transport.put_bytes(path, result)
            wt.commit(
                message=commit_message, authors=[author],
                specific_files=[path])

    def list_directory(self, directory_path):
        """Return a list of File objects for in the directory path.

        If the path doesn't exist, returns None.  If the path exists but is
        empty, an empty list is returned.  Otherwise a list of File objects in
        that directory.
        """
        if directory_path is not None:
            directory = self.get_file(directory_path)
            if directory is None or directory.file_type != FileType.DIRECTORY:
                return None
        listing = []
        wt = self.tree
        with wt.lock_read():
            for fp, fc, fkind, entry in wt.list_files(
                    from_dir=directory_path, recursive=False):
                if fc != 'V':
                    # If the file isn't versioned, skip it.
                    continue
                if directory_path is None:
                    file_path = fp
                else:
                    file_path = joinpath(directory_path, fp)
                listing.append(File(self, file_path))
            return listing


@implementer(IFile)
class File(BaseFile):
    """Represents a file in the Bazaar branch."""

    def __init__(self, filestore, path):
        BaseFile.__init__(self, path)
        self.filestore = filestore
        # This isn't entirely necessary.
        self.tree = self.filestore.tree
        self.file_type = self._get_filetype()
        self._last_modified_in_revision = None

    def _get_filetype(self):
        """Work out the filetype based on the mimetype if possible."""
        with self.tree.lock_read():
            is_directory = ('directory' == self.tree.kind(self.path))
            if is_directory:
                return FileType.DIRECTORY
            else:
                if self._mimetype is None:
                    binary = self._is_binary
                else:
                    binary = not self._mimetype.startswith('text/')
                if binary:
                    return FileType.BINARY_FILE
                else:
                    return FileType.TEXT_FILE

    def get_content(self):
        with self.tree.lock_read():
            # basis_tree is a revision tree, queries the repositry.
            # to get the stuff off the filesystem use the working tree
            # which needs to start with that.  WorkingTree.open('.').
            # branch = tree.branch.
            return self.tree.get_file_text(self.path)

    @property
    def last_modified_in_revision(self):
        if self._last_modified_in_revision is None:
            try:
                self._last_modified_in_revision = self.tree.get_file_revision(
                    self.path)
            except AttributeError:
                bt = self.tree.basis_tree()
                self._last_modified_in_revision = bt.get_file_revision(
                    self.path)
        return self._last_modified_in_revision

    @property
    def last_modified_by(self):
        """Return the first author for the revision."""
        repo = self.filestore.branch.repository
        rev = repo.get_revision(self.last_modified_in_revision)
        return rev.get_apparent_authors()[0]

    @property
    def last_modified_date(self):
        """Return the last modified date for the revision."""
        repo = self.filestore.branch.repository
        rev = repo.get_revision(self.last_modified_in_revision)
        return datetime.utcfromtimestamp(rev.timestamp)

    @property
    def _is_binary(self):
        """True if the file is binary."""
        try:
            with self.tree.lock_read():
                lines = self.tree.get_file_lines(self.path)
                check_text_lines(lines)
            return False
        except BinaryFile:
            return True

    @property
    def is_directory(self):
        """Is this file a directory?"""
        return 'directory' == self.tree.kind(self.path)

    def update(self, content, user):
        raise NotImplementedError()


class BranchFileStore(FileStore):

    def __init__(self, branch):
        self.branch = branch
        self.tree = branch.basis_tree()
        self.logger = logging.getLogger('wikkid')

    def basis_tree(self):
        return self.tree

    def update_file(self, path, content, author, parent_revision,
                    commit_message=None):
        commit_message = get_commit_message(commit_message)
        with self.branch.lock_write():
            if self.tree.is_versioned(path):
                f = File(self, path)
                content = self._get_final_text(content, f, parent_revision)
            else:
                content = normalize_content(content)
            if not isinstance(content, bytes):
                raise TypeError(content)
            with self.tree.preview_transform() as tt:
                trans_id = tt.trans_id_tree_path(path)
                if tt.tree_kind(trans_id) is not None:
                    tt.delete_contents(trans_id)
                else:
                    name = splitpath(path)[-1]
                    tt.version_file(
                        trans_id=trans_id, file_id=gen_file_id(name))
                    create_parents(tt, path, trans_id)
                tt.create_file([content], trans_id)
                try:
                    tt.commit(self.branch, commit_message, authors=[author])
                except MalformedTransform as e:
                    for conflict in e.conflicts:
                        if conflict[0] == 'non-directory parent':
                            path = FinalPaths(tt).get_path(trans_id)
                            raise FileExists(
                                '%s exists and is not a directory' %
                                conflict[1])
                    raise

            self.tree = self.branch.basis_tree()



import logging
import sys

from breezy.commands import Command
from breezy.option import Option

from breezy.workingtree import WorkingTree

from wikkid.app import WikkidApp
from wikkid.filestore.bzr import FileStore
from wikkid.user.bzr import LocalBazaarUserMiddleware

DEFAULT_PORT = 8080


def setup_logging():
    """Set up a logger sending to stderr."""
    handler = logging.StreamHandler(sys.stderr)
    fmt = '%(asctime)s %(levelname)-7s %(message)s'
    formatter = logging.Formatter(
        fmt=fmt, datefmt="%Y-%m-%d %H:%M:%S")
    handler.setFormatter(formatter)
    root = logging.getLogger()
    root.addHandler(handler)


class cmd_wikkid(Command):
    """Serve branch as a wiki using wikkid."""
    aliases = ['wiki']
    takes_args = ['branch?']
    takes_options = [
        Option(
            'port',
            help='Port to listen, defaults to 8080.',
            type=int,
            short_name='p')
        ]

    def run(self, branch=u'.', port=8080):
        setup_logging()
        logger = logging.getLogger('wikkid')
        logger.setLevel(logging.DEBUG)

        working_tree = WorkingTree.open(branch)
        logger.info('Using: %s', working_tree)
        filestore = FileStore(working_tree)

        app = WikkidApp(filestore)
        app = LocalBazaarUserMiddleware(app, working_tree.branch)
        from wsgiref.simple_server import make_server
        httpd = make_server('localhost', port, app)
        logger.info('Serving on http://localhost:%s', port)
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            logger.info('Done.')



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Tests for the view dispatcher."""

from zope.interface import Interface, implementer

from wikkid.dispatcher import get_view, register_view, unregister_view
from wikkid.tests import TestCase
from wikkid.view.base import BaseView


class TestGetView(TestCase):
    """Tests for get_view."""

    def test_no_interfaces(self):
        """If the object supports no interfaces, there is no view."""
        class NoInterfaces(object):
            pass
        obj = NoInterfaces()
        self.assertIs(None, get_view(obj, None, None))

    def test_interface_not_registered(self):
        """If the object supports an interface, but that interface is not
        registered, we get no view."""
        class IHasInterface(Interface):
            pass

        @implementer(IHasInterface)
        class HasInterface(object):
            pass
        obj = HasInterface()
        self.assertIs(None, get_view(obj, None, None))

    def test_interface_view_registered(self):
        """If the object supports an interface, and the view is registered,
        make sure that the view is returned when asked for."""
        class IHasInterface(Interface):
            pass

        @implementer(IHasInterface)
        class HasInterface(object):
            pass

        class AView(object):
            for_interface = IHasInterface
            name = 'name'

            def __init__(self, *args):
                pass

            def initialize(self):
                self.initialized = True
        register_view(AView)
        self.addCleanup(unregister_view, AView)
        obj = HasInterface()
        view = get_view(obj, 'name', None)
        self.assertIsInstance(view, AView)
        self.assertTrue(view.initialized)

    def test_interface_view_registered_default(self):
        """If the object supports an interface, and the view is registered as
        the default view, when a view is requested where no name is specified,
        the default view is returned."""
        class IHasInterface(Interface):
            pass

        @implementer(IHasInterface)
        class HasInterface(object):
            pass

        class AView(object):
            for_interface = IHasInterface
            name = 'name'
            is_default = True

            def __init__(self, *args):
                pass

            def initialize(self):
                self.initialized = True
        register_view(AView)
        self.addCleanup(unregister_view, AView)
        obj = HasInterface()
        view = get_view(obj, None, None)
        self.assertIsInstance(view, AView)
        self.assertTrue(view.initialized)


class TestViewRegistration(TestCase):
    """Test that views that inherit from BaseView are registered."""

    def test_registration(self):
        """Create a view class, and make sure it is registered."""
        class IHasInterface(Interface):
            pass

        @implementer(IHasInterface)
        class HasInterface(object):
            pass

        class AView(BaseView):
            for_interface = IHasInterface
            name = 'name'
        obj = HasInterface()
        view = get_view(obj, 'name', None)
        self.assertIsInstance(view, AView)
        self.addCleanup(unregister_view, AView)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""Model classes for wikkid.

Model objects are those objects that the views hang off.
"""



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A base test case for factory tests."""

from webob import Request

from wikkid.dispatcher import get_view
from wikkid.model.factory import ResourceFactory
from wikkid.filestore.volatile import FileStore
from wikkid.tests import TestCase


class FactoryTestCase(TestCase):
    """A test case that can make a factory."""

    def make_factory(self, content=None):
        """Make a factory with a volatile filestore."""
        filestore = FileStore(content)
        return ResourceFactory(filestore)


class ViewTestCase(FactoryTestCase):
    """A factory test case that can create views."""

    def get_view(self, factory, path, name=None, base_url=None):
        info = factory.get_resource_at_path(path)
        request = Request.blank(path, base_url=base_url)
        return get_view(info, name, request)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The model class for file resources."""

from zope.interface import implementer

from wikkid.model.baseresource import BaseResource
from wikkid.interface.resource import IFileResource, IUpdatableResource
from wikkid.user.bzr import create_bzr_user_from_author_string


@implementer(IUpdatableResource)
class UpdatableResource(BaseResource):
    """Reflects either a file either missing or actual."""

    def put_bytes(self, content: bytes, committer, rev_id, commit_msg):
        """Update the file resource."""
        self.factory.filestore.update_file(
            self.write_filename, content, committer, rev_id, commit_msg)


@implementer(IFileResource)
class FileResource(UpdatableResource):
    """Anything that relates to all files."""

    @property
    def mimetype(self):
        return self.file_resource.mimetype

    @property
    def last_modified_in_revision(self):
        return self.file_resource.last_modified_in_revision

    @property
    def last_modified_date(self):
        return self.file_resource.last_modified_date

    @property
    def last_modified_by(self):
        return create_bzr_user_from_author_string(
            self.file_resource.last_modified_by)

    def get_bytes(self):
        return self.file_resource.get_content()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""A text to html formatter using textile."""


from textile import textile
from zope.interface import implementer

from wikkid.interface.formatter import ITextFormatter


@implementer(ITextFormatter)
class TextileFormatter(object):
    """Format source files as HTML using textile."""

    def format(self, filename, text):
        """Format the text.
        """
        return textile(text)



#
# Copyright (C) 2010 Wikkid Developers.
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""The directory resource class.

A directory resource is one where the path specifies a directory in the
filestore.
"""

from zope.interface import implementer

from wikkid.model.missing import MissingResource
from wikkid.interface.filestore import FileType
from wikkid.interface.resource import IDirectoryResource


class DirectoryMethods(object):
    """Directory methods are used by DirectoryResource and WikiTextFile.

    The methods are only valid on WikiTextFile objects when there is a
    directory with the same name as the wiki file without the '.txt'
    """

    def get_dir_name(self):
        return self.dir_resource.path

    def get_listing(self):
        """Return a list of resources that are in this directory."""
        dir_name = self.get_dir_name()
        filestore = self.factory.filestore
        listing = []
        for entry in filestore.list_directory(dir_name):
            if entry.file_type == FileType.DIRECTORY:
                file_resource = None
                dir_resource = entry
            else:
                file_resource = entry
                dir_resource = None
            file_path = entry.path
            listing.append(
                self.factory.get_resource(
                    '/' + file_path, file_path, file_resource, dir_resource))

        return listing


@implementer(IDirectoryResource)
class DirectoryResource(MissingResource, DirectoryMethods):
    """A directory in the filestore.

    By definition, a directory is also a missing wiki page.
    """

    def __repr__(self):
        return "<DirectoryResource '%s'>" % self.path



#
# Copyright (C) 2010 Wikkid Developers
#
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

"""View base class for editing text pages."""

from wikkid.view.base import BaseView
from wikkid.view.utils import expand_wiki_name


class BaseEditView(BaseView):
    """Base class for editing text."""

    name = 'edit'
    template = 'edit_page'

    @property
    def title(self):
        return 'Editing "%s"' % expand_wiki_name(self.context.base_name)

    @property
    def save_url(self):
        """The link for the cancel button."""
        return self.canonical_url(self.context, 'save')

    @property
    def cancel_url(self):
        """The link for the cancel button."""
        return self.canonical_url(self.context)
